# -*- coding: utf-8 -*-
"""

"""
from typing import List

from numpy import ndarray

from pyfem.elements.BaseElement import BaseElement
from pyfem.elements.IsoElementShape import IsoElementShape
from pyfem.elements.SolidPlaneSmallStrain import SolidPlaneSmallStrain
from pyfem.elements.SolidVolumeSmallStrain import SolidVolumeSmallStrain
from pyfem.elements.ThermalStatic import ThermalStatic
from pyfem.elements.SolidThermalPlaneSmallStrain import SolidThermalPlaneSmallStrain
from pyfem.fem.Timer import Timer
from pyfem.io.Dof import Dof
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.utils.colors import error_style

element_data_dict = {
    'SolidPlaneStressSmallStrain': SolidPlaneSmallStrain,
    'SolidPlaneStrainSmallStrain': SolidPlaneSmallStrain,
    'SolidVolumeSmallStrain': SolidVolumeSmallStrain,
    'SolidThermalPlaneStrainSmallStrain': SolidThermalPlaneSmallStrain,
    'ThermalStatic': ThermalStatic
}


# @trace_calls
def get_element_data(element_id: int,
                     iso_element_shape: IsoElementShape,
                     connectivity: ndarray,
                     node_coords: ndarray,
                     dof: Dof,
                     materials: List[Material],
                     section: Section,
                     material_data_list: List[BaseMaterial],
                     timer: Timer) -> BaseElement:
    class_name = f'{section.category}{section.type}{section.option}'.strip().replace(' ', '')

    if class_name in element_data_dict:
        return element_data_dict[class_name](element_id=element_id,
                                             iso_element_shape=iso_element_shape,
                                             connectivity=connectivity,
                                             node_coords=node_coords,
                                             dof=dof,
                                             materials=materials,
                                             section=section,
                                             material_data_list=material_data_list,
                                             timer=timer)
    else:
        error_msg = f'{class_name} element is not supported.\n'
        error_msg += f'The allowed element types are {list(element_data_dict.keys())}.'
        raise NotImplementedError(error_style(error_msg))


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""

"""
from copy import deepcopy
from typing import Dict, Tuple

from numpy import zeros, ndarray, exp

from pyfem.fem.Timer import Timer
from pyfem.fem.constants import DTYPE
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.materials.ElasticIsotropic import get_stiffness_from_young_poisson
from pyfem.utils.colors import error_style


class ViscoElasticMaxwell(BaseMaterial):

    def __init__(self, material: Material, dimension: int, section: Section) -> None:
        super().__init__(material, dimension, section)
        self.allowed_section_types = ('PlaneStress', 'PlaneStrain', 'Volume')
        self.E0: float = self.material.data[0]
        self.E1: float = self.material.data[1]
        self.E2: float = self.material.data[2]
        self.E3: float = self.material.data[3]
        self.TAU1: float = self.material.data[4]
        self.TAU2: float = self.material.data[5]
        self.TAU3: float = self.material.data[6]
        self.POISSON: float = self.material.data[7]
        self.create_tangent()

    def create_tangent(self):
        young = self.material.data[0]
        poisson = self.material.data[1]

        if self.section.type in self.allowed_section_types:
            self.ddsdde = get_stiffness_from_young_poisson(self.dimension, young, poisson, self.section.type)
        else:
            error_msg = f'{self.section.type} is not the allowed section types {self.allowed_section_types} of the material {type(self).__name__}, please check the definition of the section {self.section.name}'
            raise NotImplementedError(error_style(error_msg))

    def get_tangent(self, variable: Dict[str, ndarray],
                    state_variable: Dict[str, ndarray],
                    state_variable_new: Dict[str, ndarray],
                    element_id: int,
                    igp: int,
                    ntens: int,
                    ndi: int,
                    nshr: int,
                    timer: Timer) -> Tuple[ndarray, Dict[str, ndarray]]:

        if state_variable == {}:
            state_variable['SM1'] = zeros(ntens, dtype=DTYPE)
            state_variable['SM2'] = zeros(ntens, dtype=DTYPE)
            state_variable['SM3'] = zeros(ntens, dtype=DTYPE)

        SM1OLD = deepcopy(state_variable['SM1'])
        SM2OLD = deepcopy(state_variable['SM2'])
        SM3OLD = deepcopy(state_variable['SM3'])

        dtime = timer.dtime
        dstrain = variable['dstrain']
        strain = variable['strain']

        E0 = self.E0
        E1 = self.E1
        E2 = self.E2
        E3 = self.E3
        TAU1 = self.TAU1
        TAU2 = self.TAU2
        TAU3 = self.TAU3
        POISSON = self.POISSON

        mu0 = 0.5 * E0 / (1.0 + POISSON)
        mu1 = 0.5 * E1 / (1.0 + POISSON)
        mu2 = 0.5 * E2 / (1.0 + POISSON)
        mu3 = 0.5 * E3 / (1.0 + POISSON)

        BULK = E0 / 3 / (1.0 - 2 * POISSON)

        m1 = (TAU1 * mu1 - TAU1 * mu1 * exp(-dtime / TAU1)) / (mu0 * dtime)
        m2 = (TAU2 * mu2 - TAU2 * mu2 * exp(-dtime / TAU2)) / (mu0 * dtime)
        m3 = (TAU3 * mu3 - TAU3 * mu3 * exp(-dtime / TAU3)) / (mu0 * dtime)

        term1 = BULK + (4.0 * mu0) / 3.0
        term2 = BULK - (2.0 * mu0) / 3.0

        a1 = exp(-dtime / TAU1)
        a2 = exp(-dtime / TAU2)
        a3 = exp(-dtime / TAU3)

        g = zeros((ntens, ntens), dtype=DTYPE)

        for i in range(ndi):
            g[i, i] = term1
        for i in range(1, ndi):
            for j in range(0, i):
                g[i, j] = term2
                g[j, i] = term2
        for i in range(ndi, ntens):
            g[i, i] = mu0

        stress = zeros(ntens, dtype=DTYPE)

        for i in range(ndi):
            stress[i] = g[i, 0] * strain[0] + g[i, 1] * strain[1] + g[i, 2] * strain[2] + \
                        SM1OLD[i] + SM2OLD[i] + SM3OLD[i] + (1 + m1 + m2 + m3) * \
                        (g[i, 0] * dstrain[0] + g[i, 1] * dstrain[1] + g[i, 2] * dstrain[2])

        for i in range(ndi, ntens):
            stress[i] = g[i, i] * strain[i] + SM1OLD[i] + SM2OLD[i] + SM3OLD[i] + (1 + m1 + m2 + m3) * (
                    g[i, i] * dstrain[i])

        SM1 = zeros(ntens, dtype=DTYPE)
        SM2 = zeros(ntens, dtype=DTYPE)
        SM3 = zeros(ntens, dtype=DTYPE)

        for i in range(ndi):
            SM1[i] = SM1OLD[i] + m1 * g[i, 0] * dstrain[0] + g[i, 1] * dstrain[1] + g[i, 2] * dstrain[2]
            SM2[i] = SM2OLD[i] + m2 * g[i, 0] * dstrain[0] + g[i, 1] * dstrain[1] + g[i, 2] * dstrain[2]
            SM3[i] = SM3OLD[i] + m3 * g[i, 0] * dstrain[0] + g[i, 1] * dstrain[1] + g[i, 2] * dstrain[2]

        for i in range(ndi, ntens):
            SM1[i] = SM1OLD[i] + m1 * (g[i, i] * dstrain[i])
            SM2[i] = SM2OLD[i] + m2 * (g[i, i] * dstrain[i])
            SM3[i] = SM3OLD[i] + m3 * (g[i, i] * dstrain[i])

        for i in range(ntens):
            SM1[i] = exp(-dtime / TAU1) * SM1[i]
            SM2[i] = exp(-dtime / TAU2) * SM2[i]
            SM3[i] = exp(-dtime / TAU3) * SM3[i]

        state_variable_new['SM1'] = SM1
        state_variable_new['SM2'] = SM2
        state_variable_new['SM3'] = SM3

        ddsdde = (1 + m1 + m2 + m3) * g

        output = {'stress': stress}

        return ddsdde, output


if __name__ == "__main__":
    from pyfem.Job import Job

    job = Job(r'F:\Github\pyfem\examples\specimen3D\Job-1.toml')

    material_data = ViscoElasticMaxwell(job.props.materials[2], 3, job.props.sections[0])
    print(material_data.to_string())



# -*- coding: utf-8 -*-
"""

"""
from xml.etree.ElementTree import ElementTree, Element, SubElement

from pyfem.assembly.Assembly import Assembly


def write_vtk(assembly: Assembly):
    props = assembly.props
    timer = assembly.timer
    dimension = props.mesh_data.dimension

    root = Element("VTKFile", {
        "type": "UnstructuredGrid",
        "version": "0.1",
        "byte_order": "LittleEndian"
    })

    ugrid = SubElement(root, "UnstructuredGrid")
    piece = SubElement(ugrid, "Piece", {
        "NumberOfPoints": str(len(props.mesh_data.nodes)),
        "NumberOfCells": str(len(props.mesh_data.elements))
    })

    # 添加节点数据
    point_data = SubElement(piece, "PointData")

    if "T" in props.dof.names:
        temp = SubElement(point_data, "DataArray", {
            "type": "Float64",
            "Name": "Temperature",
            "NumberOfComponents": "1",
            "format": "ascii"
        })
        temp.text = ""
        col_T = props.dof.names.index("T")
        dof_T = assembly.dof_solution.reshape(-1, len(props.dof.names))[:, col_T]
        for T in dof_T:
            temp.text += f"{T} \n"

    if "u1" in props.dof.names:
        disp = SubElement(point_data, "DataArray", {
            "type": "Float64",
            "Name": "Displacement",
            "NumberOfComponents": "3",
            "format": "ascii"
        })
        disp.text = ""
        if dimension == 2:
            for u1, u2 in assembly.dof_solution.reshape(-1, len(props.dof.names))[:, 0:2]:
                disp.text += f"{u1} {u2} 0.0 \n"
        elif dimension == 3:
            for u1, u2, u3 in assembly.dof_solution.reshape(-1, len(props.dof.names))[:, 0:3]:
                disp.text += f"{u1} {u2} {u3} \n"
        else:
            raise NotImplementedError

    for field_name, field_values in assembly.field_variables.items():
        field = SubElement(point_data, "DataArray", {
            "type": "Float64",
            "Name": field_name,
            "NumberOfComponents": "1",
            "format": "ascii"
        })
        field.text = ""
        for value in field_values:
            field.text += f"{value}\n"

    #
    points = SubElement(piece, "Points")
    node_coords = SubElement(points, "DataArray", {
        "type": "Float64",
        "NumberOfComponents": "3",
        "format": "ascii"
    })
    node_coords.text = ""
    for coord in props.mesh_data.nodes:
        if dimension == 2:
            node_coords.text += " ".join("{:.6f}".format(c) for c in coord) + " 0.0\n"
        elif dimension == 3:
            node_coords.text += " ".join("{:.6f}".format(c) for c in coord) + "\n"
        else:
            raise NotImplementedError

    #
    cells = SubElement(piece, "Cells")
    conn_elem = SubElement(cells, "DataArray", {
        "type": "Int32",
        "Name": "connectivity",
        "format": "ascii"
    })
    offset_elem = SubElement(cells, "DataArray", {
        "type": "Int32",
        "Name": "offsets",
        "format": "ascii"
    })
    types_elem = SubElement(cells, "DataArray", {
        "type": "UInt8",
        "Name": "types",
        "format": "ascii"
    })
    conn_elem.text = ""
    offset_elem.text = ""
    types_elem.text = ""
    offset = 0
    for connectivity in props.mesh_data.elements:
        conn_elem.text += " ".join(str(node_id) for node_id in connectivity) + "\n"
        offset += len(connectivity)
        offset_elem.text += "{}\n".format(offset)
        if dimension == 2:
            types_elem.text += "9\n"
        elif dimension == 3:
            types_elem.text += "12\n"  # 12表示六面体单元类型

    tree = ElementTree(root)

    job_name = props.input_file.stem

    output_file = props.work_path.joinpath(f'{job_name}-{timer.increment}.vtu')

    tree.write(output_file, xml_declaration=True, encoding='utf-8')


def write_pvd(assembly: Assembly):
    timer = assembly.timer
    job_name = assembly.props.input_file.stem
    output_file = assembly.props.work_path.joinpath(f'{job_name}.pvd')

    with open(output_file, 'w') as f:
        f.write("<VTKFile byte_order='LittleEndian' type='Collection' version='0.1'>\n")
        f.write("<Collection>\n")

        for frame in timer.frame_ids:
            f.write("<DataSet file='" + f"{job_name}-{frame}.vtu" + "' groups='' part='0' timestep='" + str(
                frame) + "'/>\n")

        f.write("</Collection>\n")
        f.write("</VTKFile>\n")


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""

"""
from typing import Optional

from pyfem.bc.BaseBC import BaseBC
from pyfem.bc.DirichletBC import DirichletBC
from pyfem.bc.NeumannBC import NeumannBC
from pyfem.bc.NeumannBCConcentrated import NeumannBCConcentrated
from pyfem.bc.NeumannBCDistributed import NeumannBCDistributed
from pyfem.io.Amplitude import Amplitude
from pyfem.io.BC import BC
from pyfem.io.Dof import Dof
from pyfem.io.Solver import Solver
from pyfem.mesh.MeshData import MeshData
from pyfem.utils.colors import error_style

bc_data_dict = {
    'DirichletBC': DirichletBC,
    'NeumannBC': NeumannBC,
    'NeumannBCConcentrated': NeumannBCConcentrated,
    'NeumannBCDistributed': NeumannBCDistributed
}


def get_bc_data(bc: BC,
                dof: Dof,
                mesh_data: MeshData,
                solver: Solver,
                amplitude: Optional[Amplitude]) -> BaseBC:
    class_name = f'{bc.category}{bc.type}'.strip().replace(' ', '')

    if class_name in bc_data_dict:
        return bc_data_dict[class_name](bc=bc,
                                        dof=dof,
                                        mesh_data=mesh_data,
                                        solver=solver,
                                        amplitude=amplitude)
    else:
        error_msg = f'{class_name} bc is not supported.\n'
        error_msg += f'The allowed bc types are {list(bc_data_dict.keys())}.'
        raise NotImplementedError(error_style(error_msg))


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""

"""
from typing import Optional

from numpy import array

from pyfem.bc.BaseBC import BaseBC
from pyfem.io.Amplitude import Amplitude
from pyfem.io.BC import BC
from pyfem.io.Dof import Dof
from pyfem.io.Solver import Solver
from pyfem.mesh.MeshData import MeshData
from pyfem.utils.colors import error_style
from pyfem.utils.visualization import object_dict_to_string_ndarray


class DirichletBC(BaseBC):
    def __init__(self, bc: BC, dof: Dof, mesh_data: MeshData, solver: Solver, amplitude: Optional[Amplitude]) -> None:
        super().__init__(bc, dof, mesh_data, solver, amplitude)
        self.create_dof_values()

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())

    def create_dof_values(self) -> None:
        bc_node_sets = self.bc.node_sets
        bc_node_ids = []
        for bc_node_set in bc_node_sets:
            bc_node_ids += list(self.mesh_data.node_sets[bc_node_set])

        # 如果发现施加当前边界条件的点集中有重复的点则抛出异常
        if len(bc_node_ids) != len(set(bc_node_ids)):
            error_msg = f'{type(self).__name__} {self.bc.name} contains repeat nodes\n'
            error_msg += f'Please check the input file'
            raise NotImplementedError(error_style(error_msg))
        else:
            self.bc_node_ids = array(bc_node_ids)

        # 确定施加的边界条件对应的全局自由度编号
        bc_dof_names = self.bc.dof
        dof_names = self.dof.names
        dof_ids = []
        for node_index in self.bc_node_ids:
            for _, bc_dof_name in enumerate(bc_dof_names):
                dof_ids.append(node_index * len(dof_names) + dof_names.index(bc_dof_name))
        self.dof_ids = array(dof_ids)

        bc_value = self.bc.value
        if isinstance(bc_value, float):
            self.dof_values = array([bc_value for _ in self.dof_ids])


if __name__ == "__main__":
    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'F:\Github\pyfem\examples\rectangle\rectangle.toml')
    props.verify()

    bc_data = DirichletBC(props.bcs[1], props.dof, props.mesh_data, props.solver, props.amplitudes[0])
    bc_data.create_dof_values()
    bc_data.show()



# -*- coding: utf-8 -*-
"""

"""
from typing import List, Callable

from scipy.interpolate import interp1d  # type: ignore

from pyfem.utils.visualization import object_dict_to_string_ndarray


class BaseAmplitude:
    def __init__(self) -> None:
        self.start: float = 0.0
        self.f_amplitude: Callable = interp1d([0, 1], [0, 1], kind='linear', fill_value='extrapolate')

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())

    def get_amplitude(self, time: float) -> float:
        return self.f_amplitude(time) + self.start

    def set_f_amplitude(self, time: List[float], value: List[float]) -> None:
        self.f_amplitude = interp1d(time, value, kind='linear', fill_value='extrapolate')


if __name__ == "__main__":
    amp = BaseAmplitude()
    amp.show()



# -*- coding: utf-8 -*-
"""

"""
from typing import Dict, List, Any


class IntKeyDict(dict):
    """
    关键字只能为整数（Integer）类型的字典，已经存在的键值无法通过赋值方法（__setitem__）更改。
    """

    def __init__(self) -> None:
        super().__init__()
        self.indices_to_ids: Dict[int, int] = {}
        self.ids_to_indices: Dict[int, int] = {}
        self.update_indices()

    def __setitem__(self, key: int, value: Any) -> None:
        if isinstance(key, int):
            if key in self:
                raise KeyError(f"Key {key} already exists")
            else:
                super().__setitem__(key, value)
        else:
            raise TypeError("Key must be an integer")

    def update_indices(self) -> None:
        list_of_keys = list(self.keys())
        self.ids_to_indices = {key: i for i, key in enumerate(list_of_keys)}
        self.indices_to_ids = {i: key for i, key in enumerate(list_of_keys)}

    def add_item_by_id(self, id_: int, item: Any) -> None:
        self[id_] = item

    def get_items_by_ids(self, ids: List[int]) -> List[Any]:
        if isinstance(ids, list):
            return [self[id_] for id_ in ids]
        else:
            raise TypeError("Argument to get_items_by_ids() must be a list")

    def get_indices_by_ids(self, ids: List[int]) -> List[int]:
        if isinstance(ids, list):
            if len(self.ids_to_indices) != len(self):
                self.update_indices()
            return [self.ids_to_indices[id] for id in ids]
        else:
            raise TypeError("Argument to get_indices_by_ids() must be a list")

    def get_ids_by_indices(self, indices: List[int]) -> List[int]:
        if isinstance(indices, list):
            if len(self.ids_to_indices) != len(self):
                self.update_indices()
            return [self.indices_to_ids[index] for index in indices]
        else:
            raise TypeError("Argument to get_ids_by_indices() must be a list")


if __name__ == "__main__":
    a = IntKeyDict()
    # a[1] = f'x{1}'
    # a[100] = f'x{100}'
    # a[10] = f'x{10}'
    for i in range(0, 1000000, 2):
        a[i] = f'x{i}'
    a.update_indices()
    print(a.get_indices_by_ids([0, 2, 4]))
    print(a.get_ids_by_indices([1, 3, 5]))



# -*- coding: utf-8 -*-
"""

"""
from typing import Tuple, Dict

from numpy import ndarray, empty

from pyfem.fem.Timer import Timer
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.utils.visualization import object_dict_to_string_ndarray


class BaseMaterial:
    def __init__(self, material: Material, dimension: int, section: Section) -> None:
        self.material: Material = material
        self.dimension: int = dimension
        self.section: Section = section
        self.allowed_section_types: Tuple = ()
        self.ddsdde: ndarray = empty(0)
        self.output: Dict[str, ndarray] = {}

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())

    def get_tangent(self, variable: Dict[str, ndarray],
                    state_variable: Dict[str, ndarray],
                    state_variable_new: Dict[str, ndarray],
                    element_id: int,
                    igp: int,
                    ntens: int,
                    ndi: int,
                    nshr: int,
                    timer: Timer) -> Tuple[ndarray, Dict[str, ndarray]]:
        return self.ddsdde, self.output



from typing import Tuple, Callable, Dict

from numpy import (empty, meshgrid, outer, column_stack, array, ndarray, insert, in1d)
from numpy.polynomial.legendre import leggauss

from pyfem.elements.IsoElementDiagram import IsoElementDiagram
from pyfem.fem.constants import DTYPE
from pyfem.utils.colors import error_style
from pyfem.utils.visualization import object_dict_to_string_ndarray


class IsoElementShape:
    """
    等参单元类，设置等参单元的形函数和积分点等信息
    """
    allowed_element_type = ['empty', 'line2', 'line3', 'tria3', 'tria6', 'quad4', 'quad8', 'tetra4', 'hex8', 'hex20']

    def __init__(self, element_type: str) -> None:
        """
        当前支持的单元类型 ['empty', 'line2', 'line3', 'tria3', 'tria6', 'quad4', 'quad8', 'tetra4', 'hex8', 'hex20']
        """
        self.element_type: str = ''
        self.diagram: str = ''
        self.dimension: int = 0
        self.nodes_number: int = 0
        self.order: int = 0
        self.shape_function: Callable = get_shape_empty
        self.gp_number = 0
        self.gp_coords: ndarray = empty(0)
        self.gp_weights: ndarray = empty(0)
        self.gp_shape_values: ndarray = empty(0)
        self.gp_shape_gradients: ndarray = empty(0)
        self.bc_surface_number: int = 0
        self.bc_surface_nodes_dict: Dict[str, Tuple] = {}
        self.bc_surface_coord_dict: Dict[str, Tuple] = {}
        self.bc_gp_coords_dict: Dict[str, ndarray] = {}
        self.bc_gp_weights: ndarray = empty(0)
        self.bc_gp_shape_values_dict: Dict[str, ndarray] = {}
        self.bc_gp_shape_gradients_dict: Dict[str, ndarray] = {}
        self.nodes_to_surface_dict: Dict[str, ndarray] = {}

        if element_type == 'empty':
            self.element_type = element_type
        elif element_type == 'line2':
            self.element_type = element_type
            self.set_line2()
        elif element_type == 'line3':
            self.element_type = element_type
            self.set_line3()
        elif element_type == 'tria3':
            self.element_type = element_type
            self.set_tria3()
        elif element_type == 'tria6':
            self.element_type = element_type
            self.set_tria6()
        elif element_type == 'quad4':
            self.element_type = element_type
            self.set_quad4()
        elif element_type == 'quad8':
            self.element_type = element_type
            self.set_quad8()
        elif element_type == 'tetra4':
            self.element_type = element_type
            self.set_tetra4()
        elif element_type == 'hex8':
            self.element_type = element_type
            self.set_hex8()
        elif element_type == 'hex20':
            self.element_type = element_type
            self.set_hex20()
        else:
            error_msg = f'Unsupported element type {element_type}'
            raise NotImplementedError(error_style(error_msg))

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())

    def set_line2(self) -> None:
        self.dimension = 1
        self.nodes_number = 2
        self.order = 1
        self.gp_coords, self.gp_weights = get_gauss_points(dimension=self.dimension, order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_line2
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.diagram = IsoElementDiagram.line2

    def set_line3(self) -> None:
        self.dimension = 1
        self.nodes_number = 3
        self.order = 2
        self.gp_coords, self.gp_weights = get_gauss_points(dimension=self.dimension, order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_line3
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.diagram = IsoElementDiagram.line3

    def set_quad4(self) -> None:
        self.dimension = 2
        self.nodes_number = 4
        self.order = 2
        self.gp_coords, self.gp_weights = get_gauss_points(dimension=self.dimension, order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_quad4
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.bc_surface_number = 4
        self.bc_surface_nodes_dict = {'s1': (3, 0),
                                      's2': (1, 2),
                                      's3': (0, 1),
                                      's4': (2, 3)}
        self.bc_surface_coord_dict = {'s1': (0, -1, 1),
                                      's2': (0, 1, 1),
                                      's3': (1, -1, 1),
                                      's4': (1, 1, 1)}
        for surface_name, surface_conn in self.bc_surface_nodes_dict.items():
            self.nodes_to_surface_dict[surface_name] = in1d(range(self.nodes_number), surface_conn)
        bc_gp_coords, self.bc_gp_weights = get_gauss_points(dimension=self.dimension - 1, order=self.order)
        self.bc_gp_coords_dict = {'s1': insert(bc_gp_coords, 0, -1, axis=1),
                                  's2': insert(bc_gp_coords, 0, 1, axis=1),
                                  's3': insert(bc_gp_coords, 1, -1, axis=1),
                                  's4': insert(bc_gp_coords, 1, 1, axis=1)}
        self.bc_gp_shape_values_dict = {}
        self.bc_gp_shape_gradients_dict = {}
        for bc_surface_name, bc_surface_gp_coords in self.bc_gp_coords_dict.items():
            bc_gp_shape_values = []
            bc_gp_shape_gradients = []
            for bc_surface_gp_coord in bc_surface_gp_coords:
                h, dhdxi = self.shape_function(bc_surface_gp_coord)
                bc_gp_shape_values.append(h)
                bc_gp_shape_gradients.append(dhdxi)
            self.bc_gp_shape_values_dict[bc_surface_name] = array(bc_gp_shape_values)
            self.bc_gp_shape_gradients_dict[bc_surface_name] = array(bc_gp_shape_gradients)
        self.diagram = IsoElementDiagram.quad4

    def set_quad8(self) -> None:
        self.dimension = 2
        self.nodes_number = 8
        self.order = 3
        self.gp_coords, self.gp_weights = get_gauss_points(dimension=self.dimension, order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_quad8
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.bc_surface_number = 4
        self.bc_surface_nodes_dict = {'s1': (3, 0, 7),
                                      's2': (1, 2, 5),
                                      's3': (0, 1, 4),
                                      's4': (2, 3, 6)}
        self.bc_surface_coord_dict = {'s1': (0, -1, 1),
                                      's2': (0, 1, 1),
                                      's3': (1, -1, 1),
                                      's4': (1, 1, 1)}
        for surface_name, surface_conn in self.bc_surface_nodes_dict.items():
            self.nodes_to_surface_dict[surface_name] = in1d(range(self.nodes_number), surface_conn)
        bc_gp_coords, self.bc_gp_weights = get_gauss_points(dimension=self.dimension - 1, order=self.order)
        self.bc_gp_coords_dict = {'s1': insert(bc_gp_coords, 0, -1, axis=1),
                                  's2': insert(bc_gp_coords, 0, 1, axis=1),
                                  's3': insert(bc_gp_coords, 1, -1, axis=1),
                                  's4': insert(bc_gp_coords, 1, 1, axis=1)}
        self.bc_gp_shape_values_dict = {}
        self.bc_gp_shape_gradients_dict = {}
        for bc_surface_name, bc_surface_gp_coords in self.bc_gp_coords_dict.items():
            bc_gp_shape_values = []
            bc_gp_shape_gradients = []
            for bc_surface_gp_coord in bc_surface_gp_coords:
                h, dhdxi = self.shape_function(bc_surface_gp_coord)
                bc_gp_shape_values.append(h)
                bc_gp_shape_gradients.append(dhdxi)
            self.bc_gp_shape_values_dict[bc_surface_name] = array(bc_gp_shape_values)
            self.bc_gp_shape_gradients_dict[bc_surface_name] = array(bc_gp_shape_gradients)
        self.diagram = IsoElementDiagram.quad8

    def set_tria3(self) -> None:
        self.dimension = 2
        self.nodes_number = 3
        self.order = 1
        self.gp_coords, self.gp_weights = get_gauss_points_triangle(order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_tria3
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.diagram = IsoElementDiagram.tria3

    def set_tria6(self) -> None:
        self.dimension = 2
        self.nodes_number = 6
        self.order = 3
        self.gp_coords, self.gp_weights = get_gauss_points_triangle(order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_tria6
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.diagram = IsoElementDiagram.tria6

    def set_tetra4(self) -> None:
        self.dimension = 3
        self.nodes_number = 4
        self.order = 1
        self.gp_coords, self.gp_weights = get_gauss_points_tetra(order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_tetra4
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.diagram = IsoElementDiagram.tetra4

    def set_hex8(self) -> None:
        self.dimension = 3
        self.nodes_number = 8
        self.order = 2
        self.gp_coords, self.gp_weights = get_gauss_points(dimension=self.dimension, order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_hex8
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.bc_surface_number = 6
        self.bc_surface_nodes_dict = {'s1': (0, 3, 7, 4),
                                      's2': (1, 2, 6, 5),
                                      's3': (0, 1, 5, 4),
                                      's4': (2, 3, 6, 7),
                                      's5': (0, 1, 2, 3),
                                      's6': (4, 5, 6, 7)}
        self.bc_surface_coord_dict = {'s1': (0, -1, 1),
                                      's2': (0, 1, 1),
                                      's3': (1, -1, 1),
                                      's4': (1, 1, 1),
                                      's5': (2, -1, 1),
                                      's6': (2, 1, 1)}
        for surface_name, surface_conn in self.bc_surface_nodes_dict.items():
            self.nodes_to_surface_dict[surface_name] = in1d(range(self.nodes_number), surface_conn)
        bc_gp_coords, self.bc_gp_weights = get_gauss_points(dimension=self.dimension - 1, order=self.order)
        self.bc_gp_coords_dict = {'s1': insert(bc_gp_coords, 0, -1, axis=1),
                                  's2': insert(bc_gp_coords, 0, 1, axis=1),
                                  's3': insert(bc_gp_coords, 1, -1, axis=1),
                                  's4': insert(bc_gp_coords, 1, 1, axis=1),
                                  's5': insert(bc_gp_coords, 2, -1, axis=1),
                                  's6': insert(bc_gp_coords, 2, 1, axis=1)}
        self.bc_gp_shape_values_dict = {}
        self.bc_gp_shape_gradients_dict = {}
        for bc_surface_name, bc_surface_gp_coords in self.bc_gp_coords_dict.items():
            bc_gp_shape_values = []
            bc_gp_shape_gradients = []
            for bc_surface_gp_coord in bc_surface_gp_coords:
                h, dhdxi = self.shape_function(bc_surface_gp_coord)
                bc_gp_shape_values.append(h)
                bc_gp_shape_gradients.append(dhdxi)
            self.bc_gp_shape_values_dict[bc_surface_name] = array(bc_gp_shape_values)
            self.bc_gp_shape_gradients_dict[bc_surface_name] = array(bc_gp_shape_gradients)
        self.diagram = IsoElementDiagram.hex8

    def set_hex20(self) -> None:
        self.dimension = 3
        self.nodes_number = 20
        self.order = 3
        self.gp_coords, self.gp_weights = get_gauss_points(dimension=self.dimension, order=self.order)
        self.gp_number = len(self.gp_weights)
        self.shape_function = get_shape_hex20
        gp_shape_values = []
        gp_shape_gradients = []
        for gp_coord in self.gp_coords:
            h, dhdxi = self.shape_function(gp_coord)
            gp_shape_values.append(h)
            gp_shape_gradients.append(dhdxi)
        self.gp_shape_values = array(gp_shape_values)
        self.gp_shape_gradients = array(gp_shape_gradients)
        self.diagram = IsoElementDiagram.hex20


def get_shape_empty(xi: ndarray) -> Tuple[ndarray, ndarray]:
    h = empty(0)
    dhdxi = empty(shape=(0, 0))

    return h, dhdxi


def get_shape_line2(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    两节点直线单元，节点序号及局部坐标方向如图所示::

        0---------------1
                +-->x0

    """
    if len(xi) != 1:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 1'))

    h = empty(2)
    dhdxi = empty(shape=(1, 2))

    h[0] = 0.5 * (1.0 - xi)
    h[1] = 0.5 * (1.0 + xi)

    dhdxi[0, 0] = -0.5
    dhdxi[0, 1] = 0.5

    return h, dhdxi


def get_shape_line3(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    三节点直线单元，节点序号及局部坐标方向如图所示::

        0-------1-------2
                +-->x0

    """
    if len(xi) != 1:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 1'))

    h = empty(3)
    dhdxi = empty(shape=(1, 3))

    h[0] = 0.5 * (1.0 - xi) - 0.5 * (1.0 - xi * xi)
    h[1] = 1 - xi * xi
    h[2] = 0.5 * (1.0 + xi) - 0.5 * (1.0 - xi * xi)

    dhdxi[0, 0] = -0.5 + xi
    dhdxi[0, 1] = -2.0 * xi
    dhdxi[0, 2] = 0.5 + xi

    return h, dhdxi


def get_shape_tria3(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    三节点三角形单元，节点序号及局部坐标方向如图所示::

        2
        * *
        *   *
        *     *
        x1      *
        |         *
        0--x0 * * * 1

    """
    if len(xi) != 2:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 2'))

    h = empty(3)
    dhdxi = empty(shape=(2, 3))
    xi = xi

    h[0] = 1.0 - xi[0] - xi[1]
    h[1] = xi[0]
    h[2] = xi[1]

    dhdxi[0, 0] = -1.0
    dhdxi[0, 1] = 1.0
    dhdxi[0, 2] = 0.0

    dhdxi[1, 0] = -1.0
    dhdxi[1, 1] = 0.0
    dhdxi[1, 2] = 1.0

    return h, dhdxi


def get_shape_quad4(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    四节点四边形单元，节点序号及局部坐标方向如图所示::

        3---------------2
        |       x1      |
        |       |       |
        |       o--x0   |
        |               |
        |               |
        0---------------1

    """
    if len(xi) != 2:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 2'))

    h = empty(4, dtype=DTYPE)
    dhdxi = empty(shape=(2, 4), dtype=DTYPE)

    h[0] = 0.25 * (1.0 - xi[0]) * (1.0 - xi[1])
    h[1] = 0.25 * (1.0 + xi[0]) * (1.0 - xi[1])
    h[2] = 0.25 * (1.0 + xi[0]) * (1.0 + xi[1])
    h[3] = 0.25 * (1.0 - xi[0]) * (1.0 + xi[1])

    dhdxi[0, 0] = -0.25 * (1.0 - xi[1])
    dhdxi[0, 1] = 0.25 * (1.0 - xi[1])
    dhdxi[0, 2] = 0.25 * (1.0 + xi[1])
    dhdxi[0, 3] = -0.25 * (1.0 + xi[1])

    dhdxi[1, 0] = -0.25 * (1.0 - xi[0])
    dhdxi[1, 1] = -0.25 * (1.0 + xi[0])
    dhdxi[1, 2] = 0.25 * (1.0 + xi[0])
    dhdxi[1, 3] = 0.25 * (1.0 - xi[0])

    return h, dhdxi


def get_shape_quad8(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    八节点四边形单元，节点序号及局部坐标方向如图所示::

        3-------6-------2
        |       x1      |
        |       |       |
        7       o--x0   5
        |               |
        |               |
        0-------4-------1

    """
    if len(xi) != 2:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 2'))

    h = empty(8)
    dhdxi = empty(shape=(2, 8))

    h[0] = -0.25 * (1.0 - xi[0]) * (1.0 - xi[1]) * (1.0 + xi[0] + xi[1])
    h[1] = -0.25 * (1.0 + xi[0]) * (1.0 - xi[1]) * (1.0 - xi[0] + xi[1])
    h[2] = -0.25 * (1.0 + xi[0]) * (1.0 + xi[1]) * (1.0 - xi[0] - xi[1])
    h[3] = -0.25 * (1.0 - xi[0]) * (1.0 + xi[1]) * (1.0 + xi[0] - xi[1])
    h[4] = 0.5 * (1.0 - xi[0]) * (1.0 + xi[0]) * (1.0 - xi[1])
    h[5] = 0.5 * (1.0 + xi[0]) * (1.0 + xi[1]) * (1.0 - xi[1])
    h[6] = 0.5 * (1.0 - xi[0]) * (1.0 + xi[0]) * (1.0 + xi[1])
    h[7] = 0.5 * (1.0 - xi[0]) * (1.0 + xi[1]) * (1.0 - xi[1])

    dhdxi[0, 0] = -0.25 * (-1.0 + xi[1]) * (2.0 * xi[0] + xi[1])
    dhdxi[0, 1] = 0.25 * (-1.0 + xi[1]) * (-2.0 * xi[0] + xi[1])
    dhdxi[0, 2] = 0.25 * (1.0 + xi[1]) * (2.0 * xi[0] + xi[1])
    dhdxi[0, 3] = -0.25 * (1.0 + xi[1]) * (-2.0 * xi[0] + xi[1])
    dhdxi[0, 4] = xi[0] * (-1.0 + xi[1])
    dhdxi[0, 5] = -0.5 * (1.0 + xi[1]) * (-1.0 + xi[1])
    dhdxi[0, 6] = -xi[0] * (1.0 + xi[1])
    dhdxi[0, 7] = 0.5 * (1.0 + xi[1]) * (-1.0 + xi[1])

    dhdxi[1, 0] = -0.25 * (-1.0 + xi[0]) * (xi[0] + 2.0 * xi[1])
    dhdxi[1, 1] = 0.25 * (1.0 + xi[0]) * (-xi[0] + 2.0 * xi[1])
    dhdxi[1, 2] = 0.25 * (1.0 + xi[0]) * (xi[0] + 2.0 * xi[1])
    dhdxi[1, 3] = -0.25 * (-1.0 + xi[0]) * (-xi[0] + 2.0 * xi[1])
    dhdxi[1, 4] = 0.5 * (1.0 + xi[0]) * (-1.0 + xi[0])
    dhdxi[1, 5] = -xi[1] * (1.0 + xi[0])
    dhdxi[1, 6] = -0.5 * (1.0 + xi[0]) * (-1.0 + xi[0])
    dhdxi[1, 7] = xi[1] * (-1.0 + xi[0])

    return h, dhdxi


def get_shape_tetra4(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    四节点四面体单元，节点序号及局部坐标方向如图所示::

        3
        * **
        *   * *
        *     *  *
        *       *   2
        *        **  *
        x2    *     * *
        |  x1         **
        0--x0 * * * * * 1

    """
    if len(xi) != 3:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 3'))

    h = empty(4)
    dhdxi = empty(shape=(3, 4))

    h[0] = 1.0 - xi[0] - xi[1] - xi[2]
    h[1] = xi[0]
    h[2] = xi[1]
    h[3] = xi[2]

    dhdxi[0, 0] = -1.0
    dhdxi[0, 1] = 1.0
    dhdxi[0, 2] = 0.0
    dhdxi[0, 3] = 0.0

    dhdxi[1, 0] = -1.0
    dhdxi[1, 1] = 0.0
    dhdxi[1, 2] = 1.0
    dhdxi[1, 3] = 0.0

    dhdxi[2, 0] = -1.0
    dhdxi[2, 1] = 0.0
    dhdxi[2, 2] = 0.0
    dhdxi[2, 3] = 1.0

    return h, dhdxi


def get_shape_hex8(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    八节点六面体单元，节点序号及局部坐标方向如图所示::

            7---------------6
           /|              /|
          / |     x2 x1   / |
         /  |     | /    /  |
        4---+-----|/----5   |
        |   |     o--x0 |   |
        |   3-----------+---2
        |  /            |  /
        | /             | /
        |/              |/
        0---------------1

    """
    if len(xi) != 3:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 3'))

    h = empty(8)
    dhdxi = empty(shape=(3, 8))

    h[0] = 0.125 * (1.0 - xi[0]) * (1.0 - xi[1]) * (1.0 - xi[2])
    h[1] = 0.125 * (1.0 + xi[0]) * (1.0 - xi[1]) * (1.0 - xi[2])
    h[2] = 0.125 * (1.0 + xi[0]) * (1.0 + xi[1]) * (1.0 - xi[2])
    h[3] = 0.125 * (1.0 - xi[0]) * (1.0 + xi[1]) * (1.0 - xi[2])
    h[4] = 0.125 * (1.0 - xi[0]) * (1.0 - xi[1]) * (1.0 + xi[2])
    h[5] = 0.125 * (1.0 + xi[0]) * (1.0 - xi[1]) * (1.0 + xi[2])
    h[6] = 0.125 * (1.0 + xi[0]) * (1.0 + xi[1]) * (1.0 + xi[2])
    h[7] = 0.125 * (1.0 - xi[0]) * (1.0 + xi[1]) * (1.0 + xi[2])

    dhdxi[0, 0] = -0.125 * (1.0 - xi[1]) * (1.0 - xi[2])
    dhdxi[0, 1] = 0.125 * (1.0 - xi[1]) * (1.0 - xi[2])
    dhdxi[0, 2] = 0.125 * (1.0 + xi[1]) * (1.0 - xi[2])
    dhdxi[0, 3] = -0.125 * (1.0 + xi[1]) * (1.0 - xi[2])
    dhdxi[0, 4] = -0.125 * (1.0 - xi[1]) * (1.0 + xi[2])
    dhdxi[0, 5] = 0.125 * (1.0 - xi[1]) * (1.0 + xi[2])
    dhdxi[0, 6] = 0.125 * (1.0 + xi[1]) * (1.0 + xi[2])
    dhdxi[0, 7] = -0.125 * (1.0 + xi[1]) * (1.0 + xi[2])
    dhdxi[1, 0] = -0.125 * (1.0 - xi[0]) * (1.0 - xi[2])

    dhdxi[1, 1] = -0.125 * (1.0 + xi[0]) * (1.0 - xi[2])
    dhdxi[1, 2] = 0.125 * (1.0 + xi[0]) * (1.0 - xi[2])
    dhdxi[1, 3] = 0.125 * (1.0 - xi[0]) * (1.0 - xi[2])
    dhdxi[1, 4] = -0.125 * (1.0 - xi[0]) * (1.0 + xi[2])
    dhdxi[1, 5] = -0.125 * (1.0 + xi[0]) * (1.0 + xi[2])
    dhdxi[1, 6] = 0.125 * (1.0 + xi[0]) * (1.0 + xi[2])
    dhdxi[1, 7] = 0.125 * (1.0 - xi[0]) * (1.0 + xi[2])

    dhdxi[2, 0] = -0.125 * (1.0 - xi[0]) * (1.0 - xi[1])
    dhdxi[2, 1] = -0.125 * (1.0 + xi[0]) * (1.0 - xi[1])
    dhdxi[2, 2] = -0.125 * (1.0 + xi[0]) * (1.0 + xi[1])
    dhdxi[2, 3] = -0.125 * (1.0 - xi[0]) * (1.0 + xi[1])
    dhdxi[2, 4] = 0.125 * (1.0 - xi[0]) * (1.0 - xi[1])
    dhdxi[2, 5] = 0.125 * (1.0 + xi[0]) * (1.0 - xi[1])
    dhdxi[2, 6] = 0.125 * (1.0 + xi[0]) * (1.0 + xi[1])
    dhdxi[2, 7] = 0.125 * (1.0 - xi[0]) * (1.0 + xi[1])

    return h, dhdxi


def get_shape_tria6(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    六节点三角形单元，节点序号及局部坐标方向如图所示::

        2
        * *
        *   *
        5     4
        x1      *
        |         *
        0--x0 3 * * 1

    """
    # u = a0 + a1 * x0 + a2 * x1 + a3 * x0 * x1 + a4 * x0 * x0 +a5 * x1 *x1
    # v = b0 + b1 * x0 + b2 * x1 + b3 * x0 * x1 + b4 * x0 * x0 +b5 * x1 *x1
    if len(xi) != 2:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 2'))

    h = empty(6)
    dhdxi = empty(shape=(2, 6))
    xi = xi

    h[0] = - xi[0] + 2.0 * xi[0] * xi[0]
    h[1] = - xi[1] + 2.0 * xi[1] * xi[1]
    h[2] = - (1.0 - xi[0] - xi[1]) + 2.0 * (1.0 - xi[0] - xi[1]) * (1.0 - xi[0] - xi[1])
    h[3] = 4.0 * xi[0] * xi[1]
    h[4] = 4.0 * xi[1] * (1.0 - xi[0] - xi[1])
    h[5] = 4.0 * xi[0] * (1.0 - xi[0] - xi[1])

    dhdxi[0, 0] = -1.0 + 4.0 * xi[0]
    dhdxi[0, 1] = 0.0
    dhdxi[0, 2] = 1.0 + 4.0 * (-1.0) * (1.0 - xi[0] - xi[1])
    dhdxi[0, 3] = 4.0 * xi[1]
    dhdxi[0, 4] = -4.0 * xi[1]
    dhdxi[0, 5] = 4.0 * (1.0 - xi[0] - xi[1]) - 4.0 * xi[0]

    dhdxi[1, 0] = 0.0
    dhdxi[1, 1] = -1.0 + 4.0 * xi[1]
    dhdxi[1, 2] = 1.0 + 4.0 * (-1.0) * (1.0 - xi[0] - xi[1])
    dhdxi[1, 3] = 4.0 * xi[0]
    dhdxi[1, 4] = 4.0 * (1.0 - xi[0] - xi[1]) - 4.0 * xi[1]
    dhdxi[1, 5] = -4.0 * xi[0]

    return h, dhdxi


def get_shape_hex20(xi: ndarray) -> Tuple[ndarray, ndarray]:
    """
    二十节点六面体单元，节点序号及局部坐标方向如图所示::

            7-------14------6
           /|              /|
         15 |     x2 x1  13 |
         /  19    | /    /  18
        4---+---12|/----5   |
        |   |     +--x0 |   |
        |   3-------10--+---2
        16 /            17 /
        |11             | 9
        |/              |/
        0-------8-------1

    """
    # u = a0 + a1 * x0 + a2 * x1 + a3 * x2 + a4 * x0 * x0 + a5 * x1 * x1 + a6 * x2 * x2 + a7 * x0 * x1 + a8 * x1 * x2 +
    #     a9 * x2 * x0 + a10 * x0^3 + a11 * x1^3 + a12 * x2^3 + a13 * x0^2 * x1 + a14 * x1^2 * x2 + a15 * x2^2 * x0 +
    #     a16 * x0 * x1^2 + a17 * x1 * x2^2 + a18 * x2 * x0^2 + a19 * x0 * x1 * x2
    # v = b0 + b1 * x0 + b2 * x1 + b3 * x2 + b4 * x0 * x0 + b5 * x1 * x1 + b6 * x2 * x2 + b7 * x0 * x1 + b8 * x1 * x2 +
    #     b9 * x2 * x0 + b10 * x0^3 + b11 * x1^3 + b12 * x2^3 + b13 * x0^2 * x1 + b14 * x1^2 * x2 + b15 * x2^2 * x0 +
    #     b16 * x0 * x1^2 + b17 * x1 * x2^2 + b18 * x2 * x0^2 + b19 * x0 * x1 * x2
    # w = c0 + c1 * x0 + c2 * x1 + c3 * x2 + c4 * x0 * x0 + c5 * x1 * x1 + c6 * x2 * x2 + c7 * x0 * x1 + c8 * x1 * x2 +
    #     c9 * x2 * x0 + c10 * x0^3 + c11 * x1^3 + c12 * x2^3 + c13 * x0^2 * x1 + c14 * x1^2 * x2 + c15 * x2^2 * x0 +
    #     c16 * x0 * x1^2 + c17 * x1 * x2^2 + c18 * x2 * x0^2 + c19 * x0 * x1 * x2
    if len(xi) != 3:
        raise NotImplementedError(error_style(f'coordinate {xi} must be dimension 3'))

    h = empty(20)
    dhdxi = empty(shape=(3, 20))

    h[0] = 0.125 * (1.0 - xi[0]) * (1.0 - xi[1]) * (1.0 - xi[2]) * (- xi[0] - xi[1] - xi[2] - 2)
    h[1] = 0.125 * (1.0 + xi[0]) * (1.0 - xi[1]) * (1.0 - xi[2]) * (xi[0] - xi[1] - xi[2] - 2)
    h[2] = 0.125 * (1.0 + xi[0]) * (1.0 + xi[1]) * (1.0 - xi[2]) * (xi[0] + xi[1] - xi[2] - 2)
    h[3] = 0.125 * (1.0 - xi[0]) * (1.0 + xi[1]) * (1.0 - xi[2]) * (- xi[0] + xi[1] - xi[2] - 2)
    h[4] = 0.125 * (1.0 - xi[0]) * (1.0 - xi[1]) * (1.0 + xi[2]) * (- xi[0] - xi[1] + xi[2] - 2)
    h[5] = 0.125 * (1.0 + xi[0]) * (1.0 - xi[1]) * (1.0 + xi[2]) * (xi[0] - xi[1] + xi[2] - 2)
    h[6] = 0.125 * (1.0 + xi[0]) * (1.0 + xi[1]) * (1.0 + xi[2]) * (xi[0] + xi[1] + xi[2] - 2)
    h[7] = 0.125 * (1.0 - xi[0]) * (1.0 + xi[1]) * (1.0 + xi[2]) * (- xi[0] + xi[1] + xi[2] - 2)
    h[8] = 0.25 * (1.0 - xi[0] * xi[0]) * (1.0 - xi[1]) * (1.0 - xi[2])
    h[9] = 0.25 * (1.0 + xi[0]) * (1.0 - xi[1] * xi[1]) * (1.0 - xi[2])
    h[10] = 0.25 * (1.0 - xi[0] * xi[0]) * (1.0 + xi[1]) * (1.0 - xi[2])
    h[11] = 0.25 * (1.0 - xi[0]) * (1.0 - xi[1] * xi[1]) * (1.0 - xi[2])
    h[12] = 0.25 * (1.0 - xi[0] * xi[0]) * (1.0 - xi[1]) * (1.0 + xi[2])
    h[13] = 0.25 * (1.0 + xi[0]) * (1.0 - xi[1] * xi[1]) * (1.0 + xi[2])
    h[14] = 0.25 * (1.0 - xi[0] * xi[0]) * (1.0 + xi[1]) * (1.0 + xi[2])
    h[15] = 0.25 * (1.0 - xi[0]) * (1.0 - xi[1] * xi[1]) * (1.0 + xi[2])
    h[16] = 0.25 * (1.0 - xi[0]) * (1.0 - xi[1]) * (1.0 - xi[2] * xi[2])
    h[17] = 0.25 * (1.0 + xi[0]) * (1.0 - xi[1]) * (1.0 - xi[2] * xi[2])
    h[18] = 0.25 * (1.0 + xi[0]) * (1.0 + xi[1]) * (1.0 - xi[2] * xi[2])
    h[19] = 0.25 * (1.0 - xi[0]) * (1.0 + xi[1]) * (1.0 - xi[2] * xi[2])

    dhdxi[0, 0] = 0.125 * (1.0 - xi[1]) * (1.0 - xi[2])
    dhdxi[0, 1] = 0.125 * (1.0 - xi[1]) * (1.0 - xi[2])
    dhdxi[0, 2] = 0.125 * (1.0 + xi[1]) * (1.0 - xi[2])
    dhdxi[0, 3] = 0.125 * (1.0 + xi[1]) * (1.0 - xi[2])
    dhdxi[0, 4] = 0.125 * (1.0 - xi[1]) * (1.0 + xi[2])
    dhdxi[0, 5] = 0.125 * (1.0 - xi[1]) * (1.0 + xi[2])
    dhdxi[0, 6] = 0.125 * (1.0 + xi[1]) * (1.0 + xi[2])
    dhdxi[0, 7] = 0.125 * (1.0 + xi[1]) * (1.0 + xi[2])
    dhdxi[0, 8] = -0.50 * xi[0] * (1.0 - xi[1]) * (1.0 - xi[2])
    dhdxi[0, 9] = 0.25 * (1.0 - xi[1] * xi[1]) * (1.0 - xi[2])
    dhdxi[0, 10] = -0.50 * xi[0] * (1.0 + xi[1]) * (1.0 - xi[2])
    dhdxi[0, 11] = -0.25 * (1.0 - xi[1] * xi[1]) * (1.0 - xi[2])
    dhdxi[0, 12] = -0.50 * xi[0] * (1.0 - xi[1]) * (1.0 + xi[2])
    dhdxi[0, 13] = 0.25 * (1.0 - xi[1] * xi[1]) * (1.0 + xi[2])
    dhdxi[0, 14] = -0.50 * xi[0] * (1.0 + xi[1]) * (1.0 + xi[2])
    dhdxi[0, 15] = -0.25 * (1.0 - xi[1] * xi[1]) * (1.0 + xi[2])
    dhdxi[0, 16] = -0.25 * (1.0 - xi[1]) * (1.0 - xi[2] * xi[2])
    dhdxi[0, 17] = 0.25 * (1.0 - xi[1]) * (1.0 - xi[2] * xi[2])
    dhdxi[0, 18] = 0.25 * (1.0 + xi[1]) * (1.0 - xi[2] * xi[2])
    dhdxi[0, 19] = -0.25 * (1.0 + xi[1]) * (1.0 - xi[2] * xi[2])

    dhdxi[1, 0] = 0.125 * (1.0 - xi[0]) * (1.0 - xi[2])
    dhdxi[1, 1] = 0.125 * (1.0 + xi[0]) * (1.0 - xi[2])
    dhdxi[1, 2] = 0.125 * (1.0 + xi[0]) * (1.0 - xi[2])
    dhdxi[1, 3] = 0.125 * (1.0 - xi[0]) * (1.0 - xi[2])
    dhdxi[1, 4] = 0.125 * (1.0 - xi[0]) * (1.0 + xi[2])
    dhdxi[1, 5] = 0.125 * (1.0 + xi[0]) * (1.0 + xi[2])
    dhdxi[1, 6] = 0.125 * (1.0 + xi[0]) * (1.0 + xi[2])
    dhdxi[1, 7] = 0.125 * (1.0 - xi[0]) * (1.0 + xi[2])
    dhdxi[1, 8] = -0.25 * (1.0 - xi[0] * xi[0]) * (1.0 - xi[2])
    dhdxi[1, 9] = -0.50 * (1.0 + xi[0]) * xi[1] * (1.0 - xi[2])
    dhdxi[1, 10] = 0.25 * (1.0 - xi[0] * xi[0]) * (1.0 - xi[2])
    dhdxi[1, 11] = -0.50 * (1.0 - xi[0]) * xi[1] * (1.0 - xi[2])
    dhdxi[1, 12] = -0.25 * (1.0 - xi[0] * xi[0]) * (1.0 + xi[2])
    dhdxi[1, 13] = -0.50 * (1.0 + xi[0]) * xi[1] * (1.0 + xi[2])
    dhdxi[1, 14] = 0.25 * (1.0 - xi[0] * xi[0]) * (1.0 + xi[2])
    dhdxi[1, 15] = -0.50 * (1.0 - xi[0]) * (1.0 + xi[2])
    dhdxi[1, 16] = -0.25 * (1.0 - xi[0]) * (1.0 - xi[2] * xi[2])
    dhdxi[1, 17] = -0.25 * (1.0 + xi[0]) * (1.0 - xi[2] * xi[2])
    dhdxi[1, 18] = 0.25 * (1.0 + xi[0]) * (1.0 - xi[2] * xi[2])
    dhdxi[1, 19] = 0.25 * (1.0 - xi[0]) * (1.0 - xi[2] * xi[2])

    dhdxi[2, 0] = 0.125 * (1.0 - xi[0]) * (1.0 - xi[1])
    dhdxi[2, 1] = 0.125 * (1.0 + xi[0]) * (1.0 - xi[1])
    dhdxi[2, 2] = 0.125 * (1.0 + xi[0]) * (1.0 + xi[1])
    dhdxi[2, 3] = 0.125 * (1.0 - xi[0]) * (1.0 + xi[1])
    dhdxi[2, 4] = 0.125 * (1.0 - xi[0]) * (1.0 - xi[1])
    dhdxi[2, 5] = 0.125 * (1.0 + xi[0]) * (1.0 - xi[1])
    dhdxi[2, 6] = 0.125 * (1.0 + xi[0]) * (1.0 + xi[1])
    dhdxi[2, 7] = 0.125 * (1.0 - xi[0]) * (1.0 + xi[1])
    dhdxi[2, 8] = -0.25 * (1.0 - xi[0] * xi[0]) * (1.0 - xi[1])
    dhdxi[2, 9] = -0.25 * (1.0 + xi[0]) * (1.0 - xi[1] * xi[1])
    dhdxi[2, 10] = -0.25 * (1.0 - xi[0] * xi[0]) * (1.0 + xi[1])
    dhdxi[2, 11] = -0.25 * (1.0 - xi[0]) * (1.0 - xi[1] * xi[1])
    dhdxi[2, 12] = 0.25 * (1.0 - xi[0] * xi[0]) * (1.0 - xi[1])
    dhdxi[2, 13] = 0.25 * (1.0 + xi[0]) * (1.0 - xi[1] * xi[1])
    dhdxi[2, 14] = 0.25 * (1.0 - xi[0] * xi[0]) * (1.0 + xi[1])
    dhdxi[2, 15] = 0.25 * (1.0 - xi[0]) * (1.0 - xi[1] * xi[1])
    dhdxi[2, 16] = -0.50 * (1.0 - xi[0]) * (1.0 - xi[1]) * xi[2]
    dhdxi[2, 17] = -0.50 * (1.0 + xi[0]) * (1.0 - xi[1]) * xi[2]
    dhdxi[2, 18] = -0.50 * (1.0 + xi[0]) * (1.0 + xi[1]) * xi[2]
    dhdxi[2, 19] = -0.50 * (1.0 - xi[0]) * (1.0 + xi[1]) * xi[2]

    return h, dhdxi


def get_gauss_points(dimension: int, order: int) -> Tuple[ndarray, ndarray]:
    xi, weight = leggauss(order)
    if dimension == 1:
        xi = xi.reshape(len(xi), -1)
        weight = weight.reshape(len(weight), -1)

    elif dimension == 2:
        xi1, xi2 = meshgrid(xi, xi)
        xi1 = xi1.ravel()
        xi2 = xi2.ravel()
        xi = column_stack((xi1, xi2))
        weight = outer(weight, weight)
        weight = weight.ravel()

    elif dimension == 3:
        xi1, xi2, xi3 = meshgrid(xi, xi, xi)
        xi1 = xi1.ravel()
        xi2 = xi2.ravel()
        xi3 = xi3.ravel()
        xi = column_stack((xi1, xi2, xi3))
        weight = outer(outer(weight, weight), weight)
        weight = weight.ravel()

    return xi.astype(DTYPE), weight.astype(DTYPE)


def get_gauss_points_triangle(order: int) -> Tuple[ndarray, ndarray]:
    if order == 1:
        xi = [[1.0 / 3.0, 1.0 / 3.0]]
        weight = [0.5]

    elif order == 3:
        r1 = 1.0 / 6.0
        r2 = 2.0 / 3.0
        xi = [[r1, r1], [r2, r1], [r1, r2]]
        w1 = 1.0 / 6.0
        weight = [w1, w1, w1]

    elif order == 7:
        r1 = 0.5 * 0.1012865073235
        r2 = 0.5 * 0.7974269853531
        r4 = 0.5 * 0.4701420641051
        r6 = 0.0597158717898
        r7 = 1.0 / 3.0
        xi = [[r1, r1], [r2, r1], [r1, r2], [r4, r6], [r4, r4], [r6, r4], [r7, r7]]
        w1 = 0.1259391805448
        w4 = 0.1323941527885
        w7 = 0.225
        weight = [w1, w1, w1, w4, w4, w4, w7]

    else:
        raise NotImplementedError(error_style('Order must be 1, 3 or 7'))

    return array(xi, dtype=DTYPE), array(weight, dtype=DTYPE)


def get_gauss_points_tetra(order: int) -> Tuple[ndarray, ndarray]:
    if order == 1:
        third = 1.0 / 3.0
        xi = [[third, third, third]]
        weight = [0.5 * third]
    else:
        raise NotImplementedError(error_style('Only order 1 integration implemented'))

    return array(xi, dtype=DTYPE), array(weight, dtype=DTYPE)


def get_gauss_points_pyramid(order: int) -> Tuple[ndarray, ndarray]:
    if order == 1:
        xi = [[0., 0., -0.5]]
        weight = [128.0 / 27.0]
    else:
        raise NotImplementedError(error_style('Only order 1 integration implemented'))

    return array(xi, dtype=DTYPE), array(weight, dtype=DTYPE)


def get_default_element_type(node_coords: ndarray) -> str:
    num_element_nodes = node_coords.shape[0]
    dimension = node_coords.shape[1]

    if dimension == 1:
        if num_element_nodes == 2:
            return "line2"
        elif num_element_nodes == 3:
            return "line3"
        else:
            error_msg = f'No 1D element with {num_element_nodes} nodes available'
            raise NotImplementedError(error_style(error_msg))
    elif dimension == 2:
        if num_element_nodes == 3:
            return "tria3"
        elif num_element_nodes == 4:
            return "quad4"
        elif num_element_nodes == 6:
            return "tria6"
        elif num_element_nodes == 8:
            return "quad8"
        elif num_element_nodes == 9:
            return "quad9"
        else:
            error_msg = f'No 2D element with {num_element_nodes} nodes available'
            raise NotImplementedError(error_style(error_msg))
    elif dimension == 3:
        if num_element_nodes == 4:
            return "tetra4"
        elif num_element_nodes == 5:
            return "pyramid5"
        elif num_element_nodes == 6:
            return "prism6"
        elif num_element_nodes == 8:
            return "hex8"
        elif num_element_nodes == 18:
            return "prism18"
        elif num_element_nodes == 20:
            return "hex20"
        else:
            error_msg = f'No 3D element with {num_element_nodes} nodes available'
            raise NotImplementedError(error_style(error_msg))
    else:
        raise NotImplementedError(error_style(f'Unsupported dimension {dimension}'))


if __name__ == "__main__":
    # iso_element_shape = IsoElementShape('tria3')
    iso_element_shape = IsoElementShape('quad4')
    # iso_element_shape = IsoElementShape('hex8')
    # iso_element_shape = IsoElementShape('quad8')
    # iso_element_shape = IsoElementShape('tetra4')
    # iso_element_shape = IsoElementShape('line2')
    # iso_element_shape = IsoElementShape('line3')
    # iso_element_shape = IsoElementShape('empty')

    # print(iso_element_shape.bc_gp_coords_dict['s1'])
    # print(iso_element_shape.bc_gp_weights)
    # print(iso_element_shape.bc_gp_shape_values_dict['s1'])
    # print(iso_element_shape.bc_gp_shape_gradients_dict['s1'])

    iso_element_shape.show()

    # print(iso_element_shape.gp_weights.dtype)



# -*- coding: utf-8 -*-
"""

"""

from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.materials.ElasticIsotropic import ElasticIsotropic
from pyfem.materials.PlasticKinematicHardening import PlasticKinematicHardening
from pyfem.materials.ThermalIsotropic import ThermalIsotropic
from pyfem.materials.ViscoElasticMaxwell import ViscoElasticMaxwell
from pyfem.utils.colors import error_style

material_data_dict = {
    'ElasticIsotropic': ElasticIsotropic,
    'PlasticKinematicHardening': PlasticKinematicHardening,
    'ViscoElasticMaxwell': ViscoElasticMaxwell,
    'ThermalIsotropic': ThermalIsotropic

}


def get_material_data(material: Material, dimension: int, section: Section) -> BaseMaterial:
    class_name = f'{material.category}{material.type}'.strip().replace(' ', '')

    if class_name in material_data_dict:
        return material_data_dict[class_name](material, dimension, section)
    else:
        error_msg = f'{class_name} material is not supported.\n'
        error_msg += f'The allowed material types are {list(material_data_dict.keys())}.'
        raise NotImplementedError(error_style(error_msg))


if __name__ == "__main__":
    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'F:\Github\pyfem\examples\rectangle\rectangle.toml')

    material_data = get_material_data(props.materials[0], 3, props.sections[0])

    print(material_data.to_string())



# -*- coding: utf-8 -*-
"""

"""
from typing import List

from numpy import array, zeros, dot, ndarray, average

from pyfem.elements.BaseElement import BaseElement
from pyfem.elements.IsoElementShape import IsoElementShape
from pyfem.fem.Timer import Timer
from pyfem.fem.constants import DTYPE
from pyfem.io.Dof import Dof
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.utils.colors import error_style


class SolidVolumeSmallStrain(BaseElement):
    __slots__ = BaseElement.__slots__ + ('gp_b_matrices', 'gp_b_matrices_transpose', 'gp_strains', 'gp_stresses')

    def __init__(self, element_id: int,
                 iso_element_shape: IsoElementShape,
                 connectivity: ndarray,
                 node_coords: ndarray,
                 dof: Dof,
                 materials: List[Material],
                 section: Section,
                 material_data_list: List[BaseMaterial],
                 timer: Timer) -> None:

        super().__init__(element_id, iso_element_shape, connectivity, node_coords)

        self.allowed_material_data_list = ('ElasticIsotropic', 'PlasticKinematicHardening', 'ViscoElasticMaxwell')
        self.allowed_material_number = 1

        self.dof = dof
        self.materials = materials
        self.section = section
        self.material_data_list = material_data_list
        self.check_materials()
        self.timer = timer

        self.dof_names = ['u1', 'u2', 'u3']
        if dof.names != self.dof_names:
            error_msg = f'{dof.names} is not the supported dof of {type(self).__name__} element'
            raise NotImplementedError(error_style(error_msg))

        element_dof_number = len(self.dof_names) * self.iso_element_shape.nodes_number
        self.element_dof_number = element_dof_number
        self.element_dof_values = zeros(element_dof_number, dtype=DTYPE)
        self.element_ddof_values = zeros(element_dof_number, dtype=DTYPE)
        self.element_fint = zeros(element_dof_number, dtype=DTYPE)
        self.element_stiffness = None  # type: ignore

        self.gp_b_matrices: ndarray = None  # type: ignore
        self.gp_b_matrices_transpose: ndarray = None  # type: ignore
        self.gp_strains: List[ndarray] = []
        self.gp_stresses: List[ndarray] = []

        self.create_gp_b_matrices()

    def create_gp_b_matrices(self) -> None:
        self.gp_b_matrices = zeros(shape=(self.iso_element_shape.gp_number, 6, self.element_dof_number))

        for igp, (gp_shape_gradient, gp_jacobi_inv) in \
                enumerate(zip(self.iso_element_shape.gp_shape_gradients, self.gp_jacobi_invs)):
            gp_dhdx = dot(gp_shape_gradient.transpose(), gp_jacobi_inv)
            for i, val in enumerate(gp_dhdx):
                self.gp_b_matrices[igp, 0, i * 3] = val[0]
                self.gp_b_matrices[igp, 1, i * 3 + 1] = val[1]
                self.gp_b_matrices[igp, 2, i * 3 + 2] = val[2]
                self.gp_b_matrices[igp, 3, i * 3] = val[1]
                self.gp_b_matrices[igp, 3, i * 3 + 1] = val[0]
                self.gp_b_matrices[igp, 4, i * 3] = val[2]
                self.gp_b_matrices[igp, 4, i * 3 + 2] = val[0]
                self.gp_b_matrices[igp, 5, i * 3 + 1] = val[2]
                self.gp_b_matrices[igp, 5, i * 3 + 2] = val[1]

        self.gp_b_matrices_transpose = array([gp_b_matrix.transpose() for gp_b_matrix in self.gp_b_matrices])

    def update_material_state(self) -> None:
        gp_number = self.iso_element_shape.gp_number
        gp_b_matrices = self.gp_b_matrices
        gp_state_variables = self.gp_state_variables
        element_dof_values = self.element_dof_values
        element_ddof_values = self.element_ddof_values
        gp_state_variables_new = self.gp_state_variables_new
        element_id = self.element_id
        timer = self.timer

        gp_ddsddes = []
        gp_strains = []
        gp_stresses = []

        for i in range(gp_number):
            gp_strain = dot(gp_b_matrices[i], element_dof_values)
            gp_dstrain = dot(gp_b_matrices[i], element_ddof_values)
            variable = {'strain': gp_strain, 'dstrain': gp_dstrain}
            gp_ddsdde, gp_output = self.material_data_list[0].get_tangent(variable=variable,
                                                                          state_variable=gp_state_variables[i],
                                                                          state_variable_new=gp_state_variables_new[i],
                                                                          element_id=element_id,
                                                                          igp=i,
                                                                          ntens=6,
                                                                          ndi=3,
                                                                          nshr=3,
                                                                          timer=timer)
            gp_stress = gp_output['stress']
            gp_ddsddes.append(gp_ddsdde)
            gp_strains.append(gp_strain)
            gp_stresses.append(gp_stress)

        self.gp_ddsddes = gp_ddsddes
        self.gp_strains = gp_strains
        self.gp_stresses = gp_stresses

    def update_element_material_stiffness_fint(self) -> None:
        element_id = self.element_id
        timer = self.timer

        gp_number = self.gp_number
        gp_b_matrices = self.gp_b_matrices
        gp_b_matrices_transpose = self.gp_b_matrices_transpose
        gp_weight_times_jacobi_dets = self.gp_weight_times_jacobi_dets

        gp_state_variables = self.gp_state_variables
        gp_state_variables_new = self.gp_state_variables_new

        element_dof_values = self.element_dof_values
        element_ddof_values = self.element_ddof_values

        self.element_stiffness = zeros(shape=(self.element_dof_number, self.element_dof_number), dtype=DTYPE)
        self.element_fint = zeros(self.element_dof_number, dtype=DTYPE)

        gp_ddsddes = []
        gp_strains = []
        gp_stresses = []

        for i in range(gp_number):
            gp_weight_times_jacobi_det = gp_weight_times_jacobi_dets[i]
            gp_b_matrix_transpose = gp_b_matrices_transpose[i]
            gp_b_matrix = gp_b_matrices[i]
            gp_strain = dot(gp_b_matrix, element_dof_values)
            gp_dstrain = dot(gp_b_matrix, element_ddof_values)
            variable = {'strain': gp_strain, 'dstrain': gp_dstrain}
            gp_ddsdde, gp_output = self.material_data_list[0].get_tangent(variable=variable,
                                                                          state_variable=gp_state_variables[i],
                                                                          state_variable_new=gp_state_variables_new[i],
                                                                          element_id=element_id,
                                                                          igp=i,
                                                                          ntens=6,
                                                                          ndi=3,
                                                                          nshr=3,
                                                                          timer=timer)
            gp_stress = gp_output['stress']

            self.element_stiffness += dot(gp_b_matrix_transpose, dot(gp_ddsdde, gp_b_matrix)) * \
                                      gp_weight_times_jacobi_det

            self.element_fint += dot(gp_b_matrix_transpose, gp_stress) * gp_weight_times_jacobi_det

            gp_ddsddes.append(gp_ddsdde)
            gp_strains.append(gp_strain)
            gp_stresses.append(gp_stress)

        self.gp_ddsddes = gp_ddsddes
        self.gp_strains = gp_strains
        self.gp_stresses = gp_stresses

    def update_element_stiffness(self) -> None:
        self.element_stiffness = zeros(shape=(self.element_dof_number, self.element_dof_number), dtype=DTYPE)

        gp_weight_times_jacobi_dets = self.gp_weight_times_jacobi_dets
        gp_b_matrices = self.gp_b_matrices
        gp_b_matrices_transpose = self.gp_b_matrices_transpose
        gp_number = self.gp_number
        gp_ddsddes = self.gp_ddsddes

        for i in range(gp_number):
            self.element_stiffness += dot(gp_b_matrices_transpose[i], dot(gp_ddsddes[i], gp_b_matrices[i])) * \
                                      gp_weight_times_jacobi_dets[i]

    def update_element_fint(self) -> None:
        gp_weight_times_jacobi_dets = self.gp_weight_times_jacobi_dets
        gp_b_matrices_transpose = self.gp_b_matrices_transpose
        gp_number = self.gp_number
        gp_stresses = self.gp_stresses

        self.element_fint = zeros(self.element_dof_number, dtype=DTYPE)
        for i in range(gp_number):
            self.element_fint += dot(gp_b_matrices_transpose[i], gp_stresses[i]) * gp_weight_times_jacobi_dets[i]

    def update_element_field_variables(self) -> None:
        gp_stresses = self.gp_stresses
        gp_strains = self.gp_strains

        average_strain = average(gp_strains, axis=0)
        average_stress = average(gp_stresses, axis=0)

        self.gp_field_variables['strain'] = array(gp_strains, dtype=DTYPE)
        self.gp_field_variables['stress'] = array(gp_stresses, dtype=DTYPE)

        self.element_average_field_variables['E11'] = average_strain[0]
        self.element_average_field_variables['E22'] = average_strain[1]
        self.element_average_field_variables['E33'] = average_strain[2]
        self.element_average_field_variables['E12'] = average_strain[3]
        self.element_average_field_variables['E13'] = average_strain[4]
        self.element_average_field_variables['E23'] = average_strain[5]
        self.element_average_field_variables['S11'] = average_stress[0]
        self.element_average_field_variables['S22'] = average_stress[1]
        self.element_average_field_variables['S33'] = average_stress[2]
        self.element_average_field_variables['S12'] = average_stress[3]
        self.element_average_field_variables['S13'] = average_stress[4]
        self.element_average_field_variables['S23'] = average_stress[5]


if __name__ == "__main__":
    from pyfem.Job import Job

    job = Job(r'F:\Github\pyfem\examples\1element\hex8\Job-1.toml')

    job.assembly.element_data_list[0].show()



# -*- coding: utf-8 -*-
"""

"""
from pyfem.utils.visualization import object_dict_to_string


class Solver:
    def __init__(self) -> None:
        self.type: str = None  # type: ignore
        self.option: str = None  # type: ignore
        self.total_time: float = 1.0
        self.start_time: float = 0.0
        self.max_increment: int = 100
        self.initial_dtime: float = 1.0
        self.max_dtime: float = 1.0
        self.min_dtime: float = 0.001
        self.dtype: str = 'float64'

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string(self, level)

    def show(self) -> None:
        print(self.to_string())


if __name__ == "__main__":
    solver = Solver()
    print(solver.__dict__.keys())
    print(solver)
    print(solver.to_string())



# -*- coding: utf-8 -*-
"""

"""
from typing import Tuple, Dict

from numpy import eye, ndarray, dot

from pyfem.fem.Timer import Timer
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.utils.colors import error_style


class ThermalIsotropic(BaseMaterial):

    def __init__(self, material: Material, dimension: int, section: Section) -> None:
        super().__init__(material, dimension, section)
        self.allowed_section_types = ('Static', 'PlaneStrain')
        self.create_tangent()

    def create_tangent(self):
        conductivity = self.material.data[0]
        capacity = self.material.data[1]

        if self.section.type in self.allowed_section_types:
            self.ddsdde = eye(self.dimension) * conductivity
        else:
            error_msg = f'{self.section.type} is not the allowed section types {self.allowed_section_types} of the material {type(self).__name__}, please check the definition of the section {self.section.name}'
            raise NotImplementedError(error_style(error_msg))

    def get_tangent(self, variable: Dict[str, ndarray],
                    state_variable: Dict[str, ndarray],
                    state_variable_new: Dict[str, ndarray],
                    element_id: int,
                    igp: int,
                    ntens: int,
                    ndi: int,
                    nshr: int,
                    timer: Timer) -> Tuple[ndarray, Dict[str, ndarray]]:
        temperature_gradient = variable['temperature_gradient']
        heat_flux = dot(-self.ddsdde, temperature_gradient)
        output = {'heat_flux': heat_flux}
        return self.ddsdde, output


if __name__ == "__main__":
    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'F:\Github\pyfem\examples\thermal\1element\hex8\Job-1.toml')

    material_data = ThermalIsotropic(props.materials[0], 3, props.sections[0])
    material_data.show()



# -*- coding: utf-8 -*-
"""

"""
from typing import List

from pyfem.utils.visualization import object_dict_to_string


class Amplitude:
    def __init__(self) -> None:
        self.name: str = None  # type: ignore
        self.type: str = None  # type: ignore
        self.start: float = 0.0
        self.data: List = None  # type: ignore

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string(self, level)

    def show(self) -> None:
        print(self.to_string())


if __name__ == "__main__":
    amp = Amplitude()
    amp.show()



# -*- coding: utf-8 -*-
"""

"""
from typing import List, Dict

from numpy import repeat, array, ndarray, empty, zeros
from scipy.sparse import coo_matrix, csc_matrix  # type: ignore

from pyfem.bc.BaseBC import BaseBC
from pyfem.bc.get_bc_data import get_bc_data
from pyfem.elements.BaseElement import BaseElement
from pyfem.elements.IsoElementShape import IsoElementShape
from pyfem.elements.get_element_data import get_element_data
from pyfem.elements.get_iso_element_type import get_iso_element_type
from pyfem.fem.Timer import Timer
from pyfem.fem.constants import DTYPE
from pyfem.io.Properties import Properties
from pyfem.materials.get_material_data import get_material_data
from pyfem.utils.colors import error_style
from pyfem.utils.visualization import object_dict_to_string_assembly
from pyfem.utils.wrappers import show_running_time

iso_element_shape_dict = {
    'line2': IsoElementShape('line2'),
    'line3': IsoElementShape('line3'),
    'tria3': IsoElementShape('tria3'),
    'tria6': IsoElementShape('tria6'),
    'quad4': IsoElementShape('quad4'),
    'quad8': IsoElementShape('quad8'),
    'tetra4': IsoElementShape('tetra4'),
    'hex8': IsoElementShape('hex8'),
    'hex20': IsoElementShape('hex20')
}


class Assembly:
    def __init__(self, props: Properties) -> None:
        self.total_dof_number: int = -1
        self.props: Properties = props
        self.timer: Timer = Timer()
        self.materials_dict: Dict = {}
        self.sections_dict: Dict = {}
        self.amplitudes_dict: Dict = {}
        self.section_of_element_set: Dict = {}
        self.element_data_list: List[BaseElement] = []
        self.bc_data_list: List[BaseBC] = []
        self.global_stiffness: csc_matrix = csc_matrix(0)
        self.fext: ndarray = empty(0, dtype=DTYPE)
        self.fint: ndarray = empty(0, dtype=DTYPE)
        self.dof_solution: ndarray = empty(0, dtype=DTYPE)
        self.ddof_solution: ndarray = empty(0, dtype=DTYPE)
        self.bc_dof_ids = empty(0)
        self.field_variables: Dict[str, ndarray] = {}
        self.init()
        self.update_element_data()
        self.assembly_global_stiffness()

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string_assembly(self, level)

    def show(self) -> None:
        print(self.to_string())

    @show_running_time
    def init(self) -> None:
        # 初始化 self.element_data_list
        mesh_data = self.props.mesh_data
        elements = mesh_data.elements
        nodes = mesh_data.nodes
        element_sets = mesh_data.element_sets
        sections = self.props.sections
        materials = self.props.materials
        solver = self.props.solver
        amplitudes = self.props.amplitudes
        dof = self.props.dof
        timer = self.timer
        dimension = self.props.mesh_data.dimension
        self.total_dof_number = len(nodes) * len(dof.names)

        for material in materials:
            self.materials_dict[material.name] = material

        for section in sections:
            self.sections_dict[section.name] = section

        for amplitude in amplitudes:
            self.amplitudes_dict[amplitude.name] = amplitude

        for element_set in element_sets:
            for section in sections:
                if element_set in section.element_sets:
                    self.section_of_element_set[element_set] = section

        for element_set_name, element_ids in element_sets.items():
            if element_set_name in self.section_of_element_set.keys():
                section = self.section_of_element_set[element_set_name]
                materials = [self.materials_dict[material_name] for material_name in section.material_names]
                material_data_list = [get_material_data(material=material,
                                                        dimension=dimension,
                                                        section=section) for material in materials]

                for element_id in element_ids:
                    connectivity = elements[element_id]
                    node_coords = nodes[connectivity]
                    iso_element_type = get_iso_element_type(node_coords)
                    iso_element_shape = iso_element_shape_dict[iso_element_type]
                    element_data = get_element_data(element_id=element_id,
                                                    iso_element_shape=iso_element_shape,
                                                    connectivity=connectivity,
                                                    node_coords=node_coords,
                                                    dof=dof,
                                                    materials=materials,
                                                    section=section,
                                                    material_data_list=material_data_list,
                                                    timer=timer)

                    element_data.assembly_conn = connectivity
                    element_data.create_element_dof_ids()

                    self.element_data_list.append(element_data)

        if len(self.element_data_list) < len(self.props.mesh_data.elements):
            raise NotImplementedError(error_style(f'some elements do not have defined section properties'))
        elif len(self.element_data_list) > len(self.props.mesh_data.elements):
            raise NotImplementedError(
                error_style(f'some elements have section properties that are redundantly defined'))

        # 初始化 self.bc_data_list
        bcs = self.props.bcs
        for bc in bcs:
            if bc.amplitude_name is not None:
                amplitude = self.amplitudes_dict[bc.amplitude_name]
            else:
                amplitude = None
            bc_data = get_bc_data(bc=bc,
                                  dof=dof,
                                  mesh_data=mesh_data,
                                  solver=solver,
                                  amplitude=amplitude)
            self.bc_data_list.append(bc_data)

        bc_dof_ids = []
        for bc_data in self.bc_data_list:
            for dof_id, dof_value in zip(bc_data.dof_ids, bc_data.dof_values):
                bc_dof_ids.append(dof_id)
        self.bc_dof_ids = array(bc_dof_ids)

        # 初始化 rhs, fext, fint, dof_solution, ddof_solution
        self.fext = zeros(self.total_dof_number, dtype=DTYPE)
        self.fint = zeros(self.total_dof_number, dtype=DTYPE)
        self.dof_solution = zeros(self.total_dof_number, dtype=DTYPE)
        self.ddof_solution = zeros(self.total_dof_number, dtype=DTYPE)

    # @show_running_time
    def assembly_global_stiffness(self) -> None:
        val = []
        row = []
        col = []

        for element_data in self.element_data_list:
            element_dof_ids = element_data.element_dof_ids
            element_dof_number = element_data.element_dof_number
            row += [r for r in repeat(element_dof_ids, element_dof_number)]
            for _ in range(element_dof_number):
                col += [c for c in element_dof_ids]
            val += [v for v in element_data.element_stiffness.reshape(element_dof_number * element_dof_number)]

        self.global_stiffness = coo_matrix(
            (array(val, dtype=DTYPE), (array(row, dtype=DTYPE), array(col, dtype=DTYPE))),
            shape=(self.total_dof_number, self.total_dof_number)).tocsc()

    # @show_running_time
    def assembly_fint(self) -> None:
        self.fint = zeros(self.total_dof_number, dtype=DTYPE)
        for element_data in self.element_data_list:
            self.fint[element_data.element_dof_ids] += element_data.element_fint

    # @show_running_time
    def update_element_data(self) -> None:
        dof_solution = self.dof_solution
        ddof_solution = self.ddof_solution
        for element_data in self.element_data_list:
            element_data.update_element_dof_values(dof_solution)
            element_data.update_element_ddof_values(ddof_solution)
            element_data.update_element_material_stiffness_fint()

    # @show_running_time
    def update_element_stiffness(self) -> None:
        for element_data in self.element_data_list:
            element_data.update_element_stiffness()

    # @show_running_time
    def update_element_data_without_stiffness(self) -> None:
        dof_solution = self.dof_solution
        ddof_solution = self.ddof_solution
        for element_data in self.element_data_list:
            element_data.update_element_dof_values(dof_solution)
            element_data.update_element_ddof_values(ddof_solution)
            element_data.update_material_state()
            element_data.update_element_fint()

    # @show_running_time
    def update_element_state_variables(self) -> None:
        for element_data in self.element_data_list:
            element_data.update_element_state_variables()

    # @show_running_time
    def update_element_field_variables(self) -> None:
        for element_data in self.element_data_list:
            element_data.update_element_field_variables()

    def assembly_field_variables(self) -> None:
        nodes_number = len(self.props.mesh_data.nodes)

        for output in self.props.outputs:
            if output.type == 'vtk':
                for field_name in output.field_outputs:
                    self.field_variables[field_name] = zeros(nodes_number)
                    nodes_count = zeros(nodes_number)
                    for element_data in self.element_data_list:
                        assembly_conn = element_data.assembly_conn
                        self.field_variables[field_name][assembly_conn] += \
                            element_data.element_average_field_variables[field_name]
                        nodes_count[assembly_conn] += 1.0
                    self.field_variables[field_name] = self.field_variables[field_name] / nodes_count


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""

"""
from typing import Optional

from pyfem.bc.BaseBC import BaseBC
from pyfem.elements.IsoElementShape import IsoElementShape
from pyfem.io.Amplitude import Amplitude
from pyfem.io.BC import BC
from pyfem.io.Dof import Dof
from pyfem.io.Solver import Solver
from pyfem.mesh.MeshData import MeshData

iso_element_shape_dict = {
    'line2': IsoElementShape('line2'),
    'line3': IsoElementShape('line3'),
    'tria3': IsoElementShape('tria3'),
    'tria6': IsoElementShape('tria6'),
    'quad4': IsoElementShape('quad4'),
    'quad8': IsoElementShape('quad8'),
    'tetra4': IsoElementShape('tetra4'),
    'hex8': IsoElementShape('hex8'),
    'hex20': IsoElementShape('hex20')
}


class NeumannBC(BaseBC):
    def __init__(self, bc: BC, dof: Dof, mesh_data: MeshData, solver: Solver, amplitude: Optional[Amplitude]) -> None:
        super().__init__(bc, dof, mesh_data, solver, amplitude)
        self.create_dof_values()

    def create_dof_values(self) -> None:
        bc_elements = self.mesh_data.bc_elements
        nodes = self.mesh_data.nodes
        dimension = self.mesh_data.dimension

        # print(self.mesh_data.elements)
        # print(self.mesh_data.bc_elements)

        bc_element_sets = self.bc.bc_element_sets

        bc_element_ids = []
        # bc_node_ids = []

        # 找到施加面力的边界单元集合

        for bc_element_set in bc_element_sets:
            bc_element_ids += list(self.mesh_data.bc_element_sets[bc_element_set])

        # print(bc_element_ids)
        #
        # for bc_element_id in bc_element_ids:
        #     print(self.mesh_data.bc_elements[bc_element_id])
        #
        # print(iso_element_shape_dict)

        # bc_element_data_list = []
        #
        # for bc_element_id in bc_element_ids:
        #     connectivity = bc_elements[bc_element_id]
        #     node_coords = nodes[connectivity]
        #     iso_element_type = get_iso_element_type(node_coords, dimension - 1)
        #     iso_element_shape = iso_element_shape_dict[iso_element_type]

        # bc_element_data_list[0].show()

        # bc_value = self.bc.value
        # if isinstance(bc_value, float):
        #     iso_element_shape = IsoElementShape('line2')  # 得到二节点线单元的等参单元
        #     gp_weights = iso_element_shape.gp_weights  # 得到高斯点权重
        #     gp_jacobi_dets = self.gp_jacobi_dets  # 得到雅可比矩阵的行列式
        #     gp_number = iso_element_shape.gp_number  # 得到高斯点数
        #     gp_shape_values = iso_element_shape.gp_shape_values
        #     gp_shape_gradients = iso_element_shape.gp_shape_gradients  # 得到等参单元形函数对坐标的导数
        #     gp_coords = iso_element_shape.gp_coords  # 得到高斯点坐标
        #     for i in range(gp_number):
        #         self.dof_values += dot(gp_shape_values[i].transpose(), bc_value[i]) * \
        #                            (sum(dot(gp_coords.transpose(), gp_shape_gradients) ** 2)) ** 0.5 * gp_weights[
        #                                i] * gp_jacobi_dets
        #     self.dof_values = array(self.dof_values)  # 定义函数在SolidPlaneSmallStrain与SolidVolumeSmallStrain中引用


if __name__ == "__main__":
    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'F:\Github\pyfem\examples\rectangle\rectangle.toml')
    # props.show()

    bc_data = NeumannBC(props.bcs[3], props.dof, props.mesh_data, props.solver, props.amplitudes[0])
    # bc_data.create_dof_values()
    # bc_data.show()

    # print(det(array([1])))



# -*- coding: utf-8 -*-
"""

"""
from copy import deepcopy
from typing import List, Dict, Tuple

from numpy import dot, array, ndarray
from numpy.linalg import det, inv

from pyfem.elements.IsoElementShape import IsoElementShape
from pyfem.fem.Timer import Timer
from pyfem.io.Dof import Dof
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.utils.colors import error_style
from pyfem.utils.visualization import object_slots_to_string_ndarray


class BaseElement:
    __slots__ = ('element_id', 'iso_element_shape', 'connectivity', 'node_coords', 'assembly_conn', 'dof', 'materials',
                 'section', 'material_data_list', 'timer', 'dof_names', 'gp_number', 'gp_jacobis', 'gp_jacobi_invs',
                 'gp_jacobi_dets', 'gp_weight_times_jacobi_dets', 'gp_ddsddes', 'gp_state_variables',
                 'gp_state_variables_new', 'gp_field_variables', 'element_dof_number', 'element_dof_ids',
                 'element_dof_values', 'element_ddof_values', 'element_fint', 'element_stiffness',
                 'element_average_field_variables', 'allowed_material_data_list', 'allowed_material_number')

    def __init__(self, element_id: int,
                 iso_element_shape: IsoElementShape,
                 connectivity: ndarray,
                 node_coords: ndarray) -> None:
        self.element_id: int = element_id  # 用户自定义的节点编号
        self.iso_element_shape: IsoElementShape = iso_element_shape
        self.connectivity: ndarray = connectivity  # 对应用户定义的节点编号
        self.node_coords: ndarray = node_coords
        self.assembly_conn: ndarray = None  # type: ignore  # 对应系统组装时的节点序号

        self.dof: Dof = None  # type: ignore
        self.materials: List[Material] = None  # type: ignore
        self.section: Section = None  # type: ignore
        self.material_data_list: List[BaseMaterial] = None  # type: ignore
        self.timer: Timer = None  # type: ignore

        self.dof_names: List[str] = []

        self.gp_number: int = self.iso_element_shape.gp_number
        self.gp_jacobis: ndarray = None  # type: ignore
        self.gp_jacobi_invs: ndarray = None  # type: ignore
        self.gp_jacobi_dets: ndarray = None  # type: ignore
        self.gp_weight_times_jacobi_dets: ndarray = None  # type: ignore
        self.gp_ddsddes: List[ndarray] = []
        self.gp_state_variables: List[Dict[str, ndarray]] = [{} for _ in range(self.gp_number)]
        self.gp_state_variables_new: List[Dict[str, ndarray]] = [{} for _ in range(self.gp_number)]
        self.gp_field_variables: Dict[str, ndarray] = {}
        self.cal_jacobi()

        self.element_dof_number: int = 0  # 单元自由度总数
        self.element_dof_ids: List[int] = []  # 对应系统组装时的自由度序号
        self.element_dof_values: ndarray = None  # type: ignore  # 对应系统组装时的自由度的值
        self.element_ddof_values: ndarray = None  # type: ignore  # 对应系统组装时的自由度增量的值
        self.element_fint: ndarray = None  # type: ignore  # 对应系统组装时的内力值
        self.element_stiffness: ndarray = None  # type: ignore
        self.element_average_field_variables: Dict[str, ndarray] = {}

        self.allowed_material_data_list: Tuple = ()
        self.allowed_material_number: int = 1

    def to_string(self, level: int = 1) -> str:
        return object_slots_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())

    def cal_jacobi(self) -> None:
        """
        通过矩阵乘法计算每个积分点上的Jacobi矩阵。
        """

        # 以下代码为采用for循环的计算方法，结构清晰，但计算效率较低
        # self.gp_jacobis = []
        # self.gp_jacobi_invs = []
        # self.gp_jacobi_dets = []
        # for gp_shape_gradient in self.iso_element_shape.gp_shape_gradients:
        #     jacobi = dot(gp_shape_gradient, self.node_coords).transpose()
        #     self.gp_jacobis.append(jacobi)
        #     self.gp_jacobi_invs.append(inv(jacobi))
        #     self.gp_jacobi_dets.append(det(jacobi))
        # self.gp_jacobis = array(self.gp_jacobis)
        # self.gp_jacobi_invs = array(self.gp_jacobi_invs)
        # self.gp_jacobi_dets = array(self.gp_jacobi_dets)

        # 以下代码为采用numpy高维矩阵乘法的计算方法，计算效率高，但要注意矩阵维度的变化
        self.gp_jacobis = dot(self.iso_element_shape.gp_shape_gradients, self.node_coords).swapaxes(1, 2)
        self.gp_jacobi_dets = det(self.gp_jacobis)

        # gp_jacobi通常为2×2或3×3的方阵，可以直接根据解析式求逆矩阵，计算效率比numpy.linalg.inv()函数更高
        self.gp_jacobi_invs = inverse(self.gp_jacobis, self.gp_jacobi_dets)
        self.gp_weight_times_jacobi_dets = self.iso_element_shape.gp_weights * self.gp_jacobi_dets

    def create_element_dof_ids(self) -> None:
        for node_index in self.assembly_conn:
            for dof_id, _ in enumerate(self.dof_names):
                self.element_dof_ids.append(node_index * len(self.dof_names) + dof_id)

    def check_materials(self) -> None:
        if len(self.materials) != self.allowed_material_number:
            error_msg = f'{type(self).__name__} section supports only {self.allowed_material_number} thermal material, please check the definition of {self.section.name}, length of {self.section.material_names} must be {self.allowed_material_number}'
            raise NotImplementedError(error_style(error_msg))
        for material_data in self.material_data_list:
            material_data_class_name = type(material_data).__name__
            if material_data_class_name not in self.allowed_material_data_list:
                error_msg = f'{material_data_class_name} is not the supported material of {type(self).__name__} section, the allowed materials are {self.allowed_material_data_list}'
                raise NotImplementedError(error_style(error_msg))

    def create_gp_b_matrices(self) -> None:
        pass

    def update_element_dof_values(self, global_dof_values: ndarray) -> None:
        self.element_dof_values = global_dof_values[self.element_dof_ids]

    def update_element_ddof_values(self, global_ddof_values: ndarray) -> None:
        self.element_ddof_values = global_ddof_values[self.element_dof_ids]

    def update_element_state_variables(self) -> None:
        self.gp_state_variables = deepcopy(self.gp_state_variables_new)

    def update_element_material_stiffness_fint(self) -> None:
        pass

    def update_material_state(self) -> None:
        pass

    def update_element_stiffness(self) -> None:
        pass

    def update_element_fint(self) -> None:
        pass

    def update_element_field_variables(self) -> None:
        pass


def inverse(gp_jacobis: ndarray, gp_jacobi_dets: ndarray) -> ndarray:
    """
    对于2×2和3×3的矩阵求逆直接带入下面的公式，其余的情况则调用np.linalg.inv()函数

    对于2×2的矩阵::

            | a11  a12 |
        A = |          |
            | a21  a22 |

        A^-1 = (1 / det(A)) * | a22  -a12 |
                              |           |
                              |-a21   a11 |

    对于3×3的矩阵::

            | a11  a12  a13 |
        A = |               |
            | a21  a22  a23 |
            |               |
            | a31  a32  a33 |

        A^-1 = (1 / det(A)) * |  A22*A33 - A23*A32   A13*A32 - A12*A33   A12*A23 - A13*A22 |
                              |                                                            |
                              |  A23*A31 - A21*A33   A11*A33 - A13*A31   A13*A21 - A11*A23 |
                              |                                                            |
                              |  A21*A32 - A22*A31   A12*A31 - A11*A32   A11*A22 - A12*A21 |


    """
    gp_jacobi_invs = []
    for A, det_A in zip(gp_jacobis, gp_jacobi_dets):
        if A.shape == (2, 2):
            gp_jacobi_invs.append(
                array([[A[1][1], -A[0][1]], [-A[1][0], A[0][0]]]) / det_A)
        elif A.shape == (3, 3):
            gp_jacobi_invs.append(array([[(A[1][1] * A[2][2] - A[1][2] * A[2][1]),
                                          (A[0][2] * A[2][1] - A[0][1] * A[2][2]),
                                          (A[0][1] * A[1][2] - A[0][2] * A[1][1])],
                                         [(A[1][2] * A[2][0] - A[1][0] * A[2][2]),
                                          (A[0][0] * A[2][2] - A[0][2] * A[2][0]),
                                          (A[0][2] * A[1][0] - A[0][0] * A[1][2])],
                                         [(A[1][0] * A[2][1] - A[1][1] * A[2][0]),
                                          (A[0][1] * A[2][0] - A[0][0] * A[2][1]),
                                          (A[0][0] * A[1][1] - A[0][1] * A[1][0])]]) / det_A)
        else:
            return inv(gp_jacobis)
    return array(gp_jacobi_invs)


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""

"""
from typing import List

from pyfem.utils.visualization import object_dict_to_string


class Timer:
    """
    计时器类，用于存储求解过程中的时间信息。
    """
    TOL_TIME = 1e-6

    def __init__(self) -> None:
        self.total_time: float = 1.0
        self.time0: float = 0.0
        self.time1: float = 0.0
        self.dtime: float = 1.0
        self.increment: int = 0
        self.frame_ids: List[int] = []

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string(self, level)

    def show(self) -> None:
        print(self.to_string())

    def is_done(self) -> bool:
        if self.time1 * (1.0 + self.TOL_TIME) >= self.total_time:
            return True
        else:
            return False


if __name__ == "__main__":
    timer = Timer()
    timer.show()



# -*- coding: utf-8 -*-
"""

"""
from numpy import empty
from scipy.sparse.linalg import spsolve  # type: ignore

from pyfem.assembly.Assembly import Assembly
from pyfem.fem.constants import DTYPE
from pyfem.io.Solver import Solver
from pyfem.io.write_vtk import write_vtk
from pyfem.solvers.BaseSolver import BaseSolver


class LinearSolver(BaseSolver):
    def __init__(self, assembly: Assembly, solver: Solver) -> None:
        super().__init__()
        self.assembly: Assembly = assembly
        self.solver: Solver = solver
        self.dof_solution = empty(0, dtype=DTYPE)
        self.PENALTY = 1.0e16

    def run(self) -> None:
        self.solve()

    def solve(self) -> None:
        A = self.assembly.global_stiffness
        rhs = self.assembly.fext

        for bc_data in self.assembly.bc_data_list:
            if bc_data.bc.category == 'DirichletBC':
                for dof_id, dof_value in zip(bc_data.dof_ids, bc_data.dof_values):
                    A[dof_id, dof_id] += self.PENALTY
                    rhs[dof_id] += dof_value * self.PENALTY
            elif bc_data.bc.category == 'NeumannBC':
                for dof_id, fext in zip(bc_data.dof_ids, bc_data.bc_fext):
                    rhs[dof_id] += fext

        x = spsolve(A, rhs)

        self.dof_solution = x
        self.assembly.dof_solution = x
        self.assembly.update_element_data()
        self.assembly.update_element_field_variables()
        self.assembly.assembly_field_variables()
        write_vtk(self.assembly)


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""

"""
from typing import Callable

from numpy import array, all, diff, ndarray
from scipy.interpolate import interp1d  # type: ignore

from pyfem.amplitude.BaseAmplitude import BaseAmplitude
from pyfem.io.Amplitude import Amplitude
from pyfem.utils.colors import error_style


class TabularAmplitude(BaseAmplitude):
    def __init__(self, amplitude: Amplitude) -> None:
        super().__init__()
        self.start = amplitude.start
        self.table: ndarray = array(amplitude.data)
        if self.table.ndim != 2:
            raise NotImplementedError(error_style('dimension of amplitude table must be 2'))
        elif self.table.shape[1] != 2:
            raise NotImplementedError(error_style('column of amplitude table must be 2'))
        elif not all(diff(self.table[:, 0]) > 0):
            raise NotImplementedError(error_style('time of amplitude table must be monotonic increase'))
        self.f_amplitude: Callable = interp1d(self.table[:, 0], self.table[:, 1], kind='linear',
                                              fill_value='extrapolate')

    def get_amplitude(self, time: float) -> float:
        return self.f_amplitude(time) + self.start


if __name__ == "__main__":
    time_table = [
        [0.0, 0.0],
        [1.0, 1.0],
        [10.0, 2.0]
    ]
    pass



# -*- coding: utf-8 -*-
"""

"""
from typing import List, Optional, Callable, Tuple

from numpy import empty, ndarray

from pyfem.amplitude.BaseAmplitude import BaseAmplitude
from pyfem.amplitude.get_amplitude_data import get_amplitude_data
from pyfem.io.Amplitude import Amplitude
from pyfem.io.BC import BC
from pyfem.io.Dof import Dof
from pyfem.io.Solver import Solver
from pyfem.mesh.MeshData import MeshData
from pyfem.utils.visualization import object_dict_to_string_ndarray


class BaseBC:
    def __init__(self, bc: BC, dof: Dof, mesh_data: MeshData, solver: Solver, amplitude: Optional[Amplitude]) -> None:
        self.bc: BC = bc
        self.dof: Dof = dof
        self.mesh_data: MeshData = mesh_data
        self.solver: Solver = solver
        self.amplitude: Optional[Amplitude] = amplitude
        if self.amplitude is not None:
            self.amplitude_data: BaseAmplitude = get_amplitude_data(self.amplitude)
        else:
            self.amplitude_data = BaseAmplitude()
            self.amplitude_data.set_f_amplitude([0, solver.total_time], [0, 1])
        self.get_amplitude: Callable = self.amplitude_data.get_amplitude
        self.bc_node_ids: ndarray = empty(0)
        self.bc_element_ids: ndarray = empty(0)
        self.dof_ids: ndarray = empty(0)
        self.dof_values: ndarray = empty(0)
        self.bc_fext: ndarray = empty(0)
        self.bc_surface: List[Tuple[int, str]] = []

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""

"""
from copy import deepcopy
from typing import Dict, Tuple

from numpy import zeros, ndarray, dot, sqrt, outer, insert, delete

from pyfem.fem.Timer import Timer
from pyfem.fem.constants import DTYPE
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.materials.ElasticIsotropic import get_stiffness_from_young_poisson
from pyfem.utils.colors import error_style


class PlasticKinematicHardening(BaseMaterial):

    def __init__(self, material: Material, dimension: int, section: Section) -> None:
        super().__init__(material, dimension, section)
        self.allowed_section_types = ('PlaneStress', 'PlaneStrain', 'Volume')
        self.young: float = self.material.data[0]
        self.poisson: float = self.material.data[1]
        self.yield_stress: float = self.material.data[2]
        self.hard: float = self.material.data[3]
        self.EBULK3: float = self.young / (1.0 - 2.0 * self.poisson)
        self.EG2: float = self.young / (1.0 + self.poisson)
        self.EG: float = self.EG2 / 2.0
        self.EG3: float = 3.0 * self.EG
        self.ELAM: float = (self.EBULK3 - self.EG2) / 3.0
        self.tolerance: float = 1.0e-10
        self.create_tangent()

    def create_tangent(self):
        if self.section.type in self.allowed_section_types:
            self.ddsdde = get_stiffness_from_young_poisson(self.dimension, self.young, self.poisson, self.section.type)
        else:
            error_msg = f'{self.section.type} is not the allowed section types {self.allowed_section_types} of the material {type(self).__name__}, please check the definition of the section {self.section.name}'
            raise NotImplementedError(error_style(error_msg))

    def get_tangent(self, variable: Dict[str, ndarray],
                    state_variable: Dict[str, ndarray],
                    state_variable_new: Dict[str, ndarray],
                    element_id: int,
                    igp: int,
                    ntens: int,
                    ndi: int,
                    nshr: int,
                    timer: Timer) -> Tuple[ndarray, Dict[str, ndarray]]:

        if state_variable == {}:
            state_variable['elastic_strain'] = zeros(ntens, dtype=DTYPE)
            state_variable['plastic_strain'] = zeros(ntens, dtype=DTYPE)
            state_variable['back_stress'] = zeros(ntens, dtype=DTYPE)
            state_variable['stress'] = zeros(ntens, dtype=DTYPE)

        elastic_strain = deepcopy(state_variable['elastic_strain'])
        plastic_strain = deepcopy(state_variable['plastic_strain'])
        back_stress = deepcopy(state_variable['back_stress'])
        stress = deepcopy(state_variable['stress'])

        dstrain = variable['dstrain']

        E = self.young
        nu = self.poisson

        if self.section.type == 'PlaneStrain':
            dstrain = insert(dstrain, 2, 0)
        elif self.section.type == 'PlaneStress':
            dstrain = insert(dstrain, 2, -nu / (1 - nu) * (dstrain[0] + dstrain[1]))

        elastic_strain += dstrain

        ddsdde = zeros((ntens, ntens), dtype=DTYPE)

        lam = E * nu / ((1.0 + nu) * (1.0 - 2.0 * nu))
        mu = E / (2.0 * (1.0 + nu))

        ddsdde[:ndi, :ndi] += lam
        for i in range(ndi):
            ddsdde[i, i] += 2 * mu
        for i in range(ndi, ntens):
            ddsdde[i, i] += mu

        stress += dot(ddsdde, dstrain)
        s = stress - back_stress
        smises = get_smises(s)

        # if element_id == 0 and igp == 0:
        #     print(ddsdde)
        #     print(dstrain)
        #     print(stress)
        #     print(dot(ddsdde, dstrain))

        if smises > (1.0 + self.tolerance) * self.yield_stress:
            hydrostatic_stress = sum(stress[:ndi]) / 3.0
            flow = stress - back_stress
            flow[:ndi] = flow[:ndi] - hydrostatic_stress
            flow *= 1.0 / smises

            delta_p = (smises - self.yield_stress) / (self.EG3 + self.hard)
            back_stress += self.hard * flow * delta_p

            plastic_strain[:ndi] += 1.5 * flow[:ndi] * delta_p
            elastic_strain[:ndi] -= 1.5 * flow[:ndi] * delta_p

            plastic_strain[ndi:] += 3.0 * flow[ndi:] * delta_p
            elastic_strain[ndi:] -= 3.0 * flow[ndi:] * delta_p

            stress = back_stress + flow * self.yield_stress
            stress[:ndi] += hydrostatic_stress

            # if element_id == 0 and igp == 0:
            #     print(stress)

            EFFG = self.EG * (self.yield_stress + self.hard * delta_p) / smises
            EFFG2 = 2.0 * EFFG
            EFFG3 = 3.0 * EFFG
            EFFLAM = (self.EBULK3 - EFFG2) / 3.0
            EFFHRD = self.EG3 * self.hard / (self.EG3 + self.hard) - EFFG3

            ddsdde = zeros(shape=(ntens, ntens), dtype=DTYPE)
            ddsdde[:ndi, :ndi] = EFFLAM

            for i in range(ndi):
                ddsdde[i, i] += EFFG2

            for i in range(ndi, ntens):
                ddsdde[i, i] += EFFG

            ddsdde += EFFHRD * outer(flow, flow)

        # if element_id == 0 and igp == 0:
        #     print(stress)

        state_variable_new['elastic_strain'] = elastic_strain
        state_variable_new['plastic_strain'] = plastic_strain
        state_variable_new['back_stress'] = back_stress
        state_variable_new['stress'] = stress

        if self.section.type == 'PlaneStrain':
            ddsdde = delete(delete(ddsdde, 2, axis=0), 2, axis=1)
            stress = delete(stress, 2)
        elif self.section.type == 'PlaneStress':
            ddsdde = delete(delete(ddsdde, 2, axis=0), 2, axis=1)
            ddsdde[0, 0] -= lam * lam / (lam + 2 * mu)
            ddsdde[0, 1] -= lam * lam / (lam + 2 * mu)
            ddsdde[1, 0] -= lam * lam / (lam + 2 * mu)
            ddsdde[1, 1] -= lam * lam / (lam + 2 * mu)
            stress = delete(stress, 2)

        output = {'stress': stress}

        return ddsdde, output


def get_smises(s: ndarray) -> float:
    if len(s) == 3:
        smises = sqrt(s[0] ** 2 + s[1] ** 2 - s[0] * s[1] + 3 * s[2] ** 2)
        return float(smises)
    elif len(s) >= 3:
        smises = (s[0] - s[1]) ** 2 + (s[1] - s[2]) ** 2 + (s[2] - s[0]) ** 2
        smises += 6 * sum([i ** 2 for i in s[3:]])
        smises = sqrt(0.5 * smises)
        return float(smises)
    else:
        raise NotImplementedError(error_style(f'unsupported stress dimension {len(s)}'))


if __name__ == "__main__":
    from pyfem.Job import Job

    job = Job(r'F:\Github\pyfem\examples\rectangle\rectangle.toml')

    material_data = PlasticKinematicHardening(job.props.materials[0], 3, job.props.sections[0])



# -*- coding: utf-8 -*-
"""

"""
from pyfem.amplitude.BaseAmplitude import BaseAmplitude
from pyfem.amplitude.TabularAmplitude import TabularAmplitude
from pyfem.io.Amplitude import Amplitude
from pyfem.utils.colors import error_style

amplitude_data_dict = {
    'TabularAmplitude': TabularAmplitude
}


def get_amplitude_data(amplitude: Amplitude) -> BaseAmplitude:
    class_name = f'{amplitude.type}'.strip().replace(' ', '')

    if class_name in amplitude_data_dict:
        return amplitude_data_dict[class_name](amplitude=amplitude)
    else:
        error_msg = f'{class_name} bc is not supported.\n'
        error_msg += f'The allowed bc types are {list(amplitude_data_dict.keys())}.'
        raise NotImplementedError(error_style(error_msg))


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""

"""
from pyfem.assembly.Assembly import Assembly
from pyfem.io.Solver import Solver
from pyfem.solvers.BaseSolver import BaseSolver
from pyfem.solvers.LinearSolver import LinearSolver
from pyfem.solvers.NonlinearSolver import NonlinearSolver
from pyfem.utils.colors import error_style

solver_data_dict = {
    'LinearSolver': LinearSolver,
    'NonlinearSolver': NonlinearSolver
}


def get_solver_data(assembly: Assembly, solver: Solver) -> BaseSolver:
    class_name = f'{solver.type}'.strip().replace(' ', '')

    if class_name in solver_data_dict:
        return solver_data_dict[class_name](assembly=assembly,
                                            solver=solver)
    else:
        error_msg = f'{class_name} solver is not supported.\n'
        error_msg += f'The allowed solver types are {list(solver_data_dict.keys())}.'
        raise NotImplementedError(error_style(error_msg))


if __name__ == "__main__":
    pass



from numpy import ndarray

from pyfem.utils.colors import error_style


def get_iso_element_type(node_coords: ndarray, dimension: int = -1) -> str:
    element_node_number = node_coords.shape[0]
    if dimension == -1:
        dimension = node_coords.shape[1]

    if dimension == 1:
        if element_node_number == 2:
            return 'line2'
        elif element_node_number == 3:
            return 'line3'
        else:
            error_msg = f'no 1D element with {element_node_number} nodes available'
            raise NotImplementedError(error_style(error_msg))
    elif dimension == 2:
        if element_node_number == 3:
            return 'tria3'
        elif element_node_number == 4:
            return 'quad4'
        elif element_node_number == 6:
            return 'tria6'
        elif element_node_number == 8:
            return 'quad8'
        elif element_node_number == 9:
            return 'quad9'
        else:
            error_msg = f'no 2D element with {element_node_number} nodes available'
            raise NotImplementedError(error_style(error_msg))
    elif dimension == 3:
        if element_node_number == 4:
            return 'tetra4'
        elif element_node_number == 5:
            return 'pyramid5'
        elif element_node_number == 6:
            return 'prism6'
        elif element_node_number == 8:
            return 'hex8'
        elif element_node_number == 18:
            return 'prism18'
        elif element_node_number == 20:
            return 'hex20'
        else:
            error_msg = f'no 3D element with {element_node_number} nodes available'
            raise NotImplementedError(error_style(error_msg))
    else:
        error_msg = 'Dimension must be 1,2 or 3'
        raise NotImplementedError(error_style(error_msg))



from numpy import ndarray, empty

from pyfem.assembly.Assembly import Assembly
from pyfem.io.Solver import Solver
from pyfem.utils.visualization import object_dict_to_string_ndarray


class BaseSolver:
    def __init__(self) -> None:
        self.assembly: Assembly = None  # type: ignore
        self.solver: Solver = None  # type: ignore
        self.solution: ndarray = empty(0)

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())

    def run(self) -> None:
        pass

    def solve(self) -> None:
        pass


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""

"""
from typing import Optional, List

from pyfem.utils.visualization import object_dict_to_string


class Section:
    def __init__(self) -> None:
        self.name: str = None  # type: ignore
        self.category: str = None  # type: ignore
        self.type: str = None  # type: ignore
        self.option: Optional[str] = None
        self.element_sets: List[str] = None  # type: ignore
        self.material_names: List[str] = None  # type: ignore
        self.data: List = None  # type: ignore

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string(self, level)

    def show(self) -> None:
        print(self.to_string())


if __name__ == "__main__":
    section = Section()
    print(section.__dict__.keys())
    print(section)
    print(section.to_string())



# -*- coding: utf-8 -*-
"""

"""
from pyfem.utils.visualization import object_dict_to_string


class Output:
    def __init__(self) -> None:
        self.type: str = None  # type: ignore
        self.field_outputs: str = None  # type: ignore
        self.on_screen: bool = None  # type: ignore

    def to_string(self, level: int = 1) -> str:
        return object_dict_to_string(self, level)

    def show(self) -> None:
        print(self.to_string())


if __name__ == "__main__":
    output = Output()
    print(output.__dict__.keys())
    print(output)
    print(output.to_string())



# -*- coding: utf-8 -*-
"""

"""
from typing import Optional, Tuple, Dict

from numpy import array, outer, diag, ndarray, dot

from pyfem.fem.Timer import Timer
from pyfem.fem.constants import DTYPE
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.utils.colors import error_style


class ElasticIsotropic(BaseMaterial):

    def __init__(self, material: Material, dimension: int, section: Section) -> None:
        super().__init__(material, dimension, section)
        self.allowed_section_types = ('Volume', 'PlaneStress', 'PlaneStrain', None)
        self.create_tangent()

    def create_tangent(self):
        young = self.material.data[0]
        poisson = self.material.data[1]

        if self.section.type in self.allowed_section_types:
            self.ddsdde = get_stiffness_from_young_poisson(self.dimension, young, poisson, self.section.type)
        else:
            error_msg = f'{self.section.type} is not the allowed section types {self.allowed_section_types} of the material {type(self).__name__}, please check the definition of the section {self.section.name}'
            raise NotImplementedError(error_style(error_msg))

    def get_tangent(self, variable: Dict[str, ndarray],
                    state_variable: Dict[str, ndarray],
                    state_variable_new: Dict[str, ndarray],
                    element_id: int,
                    igp: int,
                    ntens: int,
                    ndi: int,
                    nshr: int,
                    timer: Timer) -> Tuple[ndarray, Dict[str, ndarray]]:
        strain = variable['strain']
        stress = dot(self.ddsdde, strain)
        output = {'stress': stress}
        return self.ddsdde, output


def get_lame_from_young_poisson(young: float, poisson: float, plane: Optional[str]) -> Tuple[float, float]:
    r"""
    Compute Lamé parameters from Young's modulus and Poisson's ratio.

    The relationship between Lamé parameters and Young's modulus, Poisson's
    ratio (see [1],[2]):

    .. math::
        \lambda = {\nu E \over (1+\nu)(1-2\nu)},\qquad \mu = {E \over 2(1+\nu)}

    The plain stress hypothesis:

    .. math::
       \bar\lambda = {2\lambda\mu \over \lambda + 2\mu}

    [1] I.S. Sokolnikoff: Mathematical Theory of Elasticity. New York, 1956.

    [2] T.J.R. Hughes: The Finite Element Method, Linear Static and Dynamic
    Finite Element Analysis. New Jersey, 1987.
    """
    mu = young / (2.0 * (1.0 + poisson))
    lam = young * poisson / ((1.0 + poisson) * (1.0 - 2.0 * poisson))

    if plane == 'PlaneStress':
        lam = 2 * lam * mu / (lam + 2 * mu)

    return lam, mu


def get_stiffness_from_lame(dim: int, lam: float, mu: float) -> ndarray:
    r"""
    Compute stiffness tensor corresponding to Lamé parameters.

    .. math::
        {\bf D}_{(2D)} = \begin{bmatrix} \lambda + 2\mu & \lambda & 0\\
        \lambda & \lambda + 2\mu & 0\\ 0 & 0 & \mu \end{bmatrix}

    .. math::
        {\bf D}_{(3D)} = \begin{bmatrix} \lambda + 2\mu & \lambda &
        \lambda & 0 & 0 & 0\\ \lambda & \lambda + 2\mu & \lambda & 0 & 0 & 0 \\
        \lambda & \lambda & \lambda + 2\mu & 0 & 0 & 0 \\ 0 & 0 & 0 & \mu & 0 &
        0 \\ 0 & 0 & 0 & 0 & \mu & 0 \\ 0 & 0 & 0 & 0 & 0 & \mu\\ \end{bmatrix}
    """
    sym = (dim + 1) * dim // 2
    o = array([1.] * dim + [0.] * (sym - dim), dtype=DTYPE)
    oot = outer(o, o)
    do1 = diag(o + 1.0)

    lam_array = array(lam, dtype=DTYPE)[..., None, None]
    mu_array = array(mu, dtype=DTYPE)[..., None, None]
    return lam_array * oot + mu_array * do1


def get_stiffness_from_young_poisson(dim: int, young: float, poisson: float, plane: Optional[str]) -> ndarray:
    """
    Compute stiffness tensor corresponding to Young's modulus and Poisson's
    ratio.
    """

    lam, mu = get_lame_from_young_poisson(young, poisson, plane)

    return get_stiffness_from_lame(dim, lam, mu)


def get_stiffness_from_lame_mixed(dim: int, lam: float, mu: float) -> ndarray:
    r"""
    Compute stiffness tensor corresponding to Lamé parameters for mixed
    formulation.

    .. math::
        {\bf D}_{(2D)} = \begin{bmatrix} \widetilde\lambda + 2\mu &
        \widetilde\lambda & 0\\ \widetilde\lambda & \widetilde\lambda + 2\mu &
        0\\ 0 & 0 & \mu \end{bmatrix}

    .. math::
        {\bf D}_{(3D)} = \begin{bmatrix} \widetilde\lambda + 2\mu &
        \widetilde\lambda & \widetilde\lambda & 0 & 0 & 0\\ \widetilde\lambda &
        \widetilde\lambda + 2\mu & \widetilde\lambda & 0 & 0 & 0 \\
        \widetilde\lambda & \widetilde\lambda & \widetilde\lambda + 2\mu & 0 &
        0 & 0 \\ 0 & 0 & 0 & \mu & 0 & 0 \\ 0 & 0 & 0 & 0 & \mu & 0 \\ 0 & 0 &
        0 & 0 & 0 & \mu\\ \end{bmatrix}

    where

    .. math::
       \widetilde\lambda = -{2\over 3} \mu
    """
    lam = - 2.0 / 3.0 * mu

    return get_stiffness_from_lame(dim, lam, mu)


def get_stiffness_from_young_poisson_mixed(dim: int, young: float, poisson: float, plane) -> ndarray:
    """
    Compute stiffness tensor corresponding to Young's modulus and Poisson's
    ratio for mixed formulation.
    """
    lam, mu = get_lame_from_young_poisson(young, poisson, plane)

    return get_stiffness_from_lame_mixed(dim, lam, mu)


def get_bulk_from_lame(lam: float, mu: float) -> float:
    r"""
    Compute bulk modulus from Lamé parameters.

    .. math::
        \gamma = \lambda + {2 \over 3} \mu
    """
    return lam + 2.0 * mu / 3.0


def get_bulk_from_young_poisson(young: float, poisson: float, plane: Optional[str]) -> float:
    """
    Compute bulk modulus corresponding to Young's modulus and Poisson's ratio.
    """
    lam, mu = get_lame_from_young_poisson(young, poisson, plane)

    return get_bulk_from_lame(lam, mu)


def get_lame_from_stiffness(stiffness: ndarray, plane: Optional[str]) -> Tuple[float, float]:
    """
    Compute Lamé parameters from an isotropic stiffness tensor.
    """
    lam = float(stiffness[..., 0, 1])
    mu = float(stiffness[..., -1, -1])
    if plane == 'PlaneStress':
        lam = - 2.0 * mu * lam / (lam - 2.0 * mu)

    return lam, mu


def get_young_poisson_from_stiffness(stiffness: ndarray, plane: Optional[str]) -> Tuple[float, float]:
    """
    Compute Young's modulus and Poisson's ratio from an isotropic stiffness
    tensor.
    """
    lam, mu = get_lame_from_stiffness(stiffness, plane)
    young = (3 * lam * mu + 2 * mu ** 2) / (lam + mu)
    poisson = lam / (2 * lam + 2 * mu)

    return young, poisson


if __name__ == "__main__":
    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'F:\Github\pyfem\examples\rectangle\rectangle.toml')

    material_data = ElasticIsotropic(props.materials[0], 3, props.sections[0])
    print(material_data.ddsdde.dtype)



# -*- coding: utf-8 -*-
"""

"""
from typing import Dict, List, Optional, Tuple

from numpy import array, delete, dot, logical_and, ndarray, in1d, all, sqrt, zeros
from numpy.linalg import det

from pyfem.bc.BaseBC import BaseBC
from pyfem.elements.IsoElementShape import IsoElementShape
from pyfem.elements.get_iso_element_type import get_iso_element_type
from pyfem.io.Amplitude import Amplitude
from pyfem.io.BC import BC
from pyfem.io.Dof import Dof
from pyfem.io.Solver import Solver
from pyfem.mesh.MeshData import MeshData
from pyfem.utils.colors import error_style

iso_element_shape_dict: Dict[str, IsoElementShape] = {
    'line2': IsoElementShape('line2'),
    'line3': IsoElementShape('line3'),
    'tria3': IsoElementShape('tria3'),
    'tria6': IsoElementShape('tria6'),
    'quad4': IsoElementShape('quad4'),
    'quad8': IsoElementShape('quad8'),
    'tetra4': IsoElementShape('tetra4'),
    'hex8': IsoElementShape('hex8'),
    'hex20': IsoElementShape('hex20')
}


class NeumannBCDistributed(BaseBC):
    def __init__(self, bc: BC, dof: Dof, mesh_data: MeshData, solver: Solver, amplitude: Optional[Amplitude]) -> None:
        super().__init__(bc, dof, mesh_data, solver, amplitude)
        self.create_dof_values()

    def get_surface_from_bc_element(self, bc_element_id: int, bc_element: ndarray) -> List[Tuple[int, str]]:
        nodes = self.mesh_data.nodes
        elements = self.mesh_data.elements
        element_surface = []
        for element_id, element in enumerate(elements):
            is_element_surface = all(in1d(bc_element, element))
            if is_element_surface:
                nodes_in_element = in1d(element, bc_element)
                connectivity = elements[element_id]
                node_coords = nodes[connectivity]
                iso_element_type = get_iso_element_type(node_coords)
                iso_element_shape = iso_element_shape_dict[iso_element_type]
                surface_names = [surface_name for surface_name, nodes_on_surface in
                                 iso_element_shape.nodes_to_surface_dict.items() if
                                 all(nodes_on_surface == nodes_in_element)]
                if len(surface_names) == 1:
                    element_surface.append((element_id, surface_names[0]))
                else:
                    raise NotImplementedError(error_style(f'the surface of element {element_id} is wrong'))

        if len(element_surface) == 1:
            return element_surface
        else:
            raise NotImplementedError(error_style(f'the surface of bc_element {bc_element_id} is wrong'))

    def get_surface_from_elements_nodes(self, element_id: int, node_ids: List[int]) -> List[Tuple[int, str]]:
        nodes = self.mesh_data.nodes
        elements = self.mesh_data.elements
        element_surface = []
        nodes_in_element = in1d(elements[element_id], node_ids)
        connectivity = elements[element_id]
        node_coords = nodes[connectivity]
        iso_element_type = get_iso_element_type(node_coords)
        iso_element_shape = iso_element_shape_dict[iso_element_type]
        surface_names = [surface_name for surface_name, nodes_on_surface in
                         iso_element_shape.nodes_to_surface_dict.items() if
                         sum(logical_and(nodes_in_element, nodes_on_surface)) == len(
                             iso_element_shape.bc_surface_nodes_dict[surface_name])]

        for surface_name in surface_names:
            element_surface.append((element_id, surface_name))

        if 1 <= len(element_surface) <= iso_element_shape.bc_surface_number:
            return element_surface
        else:
            raise NotImplementedError(error_style(f'the surface of element {element_id} is wrong'))

    def create_dof_values(self) -> None:
        dimension = self.mesh_data.dimension
        nodes = self.mesh_data.nodes
        elements = self.mesh_data.elements
        bc_elements = self.mesh_data.bc_elements

        node_sets = self.bc.node_sets
        element_sets = self.bc.element_sets
        bc_element_sets = self.bc.bc_element_sets
        bc_value = self.bc.value

        if bc_element_sets is not None:
            bc_element_ids = []
            for bc_element_set in bc_element_sets:
                bc_element_ids += list(self.mesh_data.bc_element_sets[bc_element_set])
            for bc_element_id in set(bc_element_ids):
                self.bc_surface += self.get_surface_from_bc_element(bc_element_id, bc_elements[bc_element_id])
        elif element_sets is not None and node_sets is not None:
            if element_sets == node_sets:
                element_ids = []
                for element_set in element_sets:
                    element_ids += list(self.mesh_data.element_sets[element_set])
                node_ids = []
                for node_set in node_sets:
                    node_ids += list(self.mesh_data.node_sets[node_set])
                for element_id in set(element_ids):
                    self.bc_surface += self.get_surface_from_elements_nodes(element_id, node_ids)
            else:
                raise NotImplementedError(
                    error_style(f'the name of element_sets {element_sets} and node_sets {node_sets} must be the same'))

        dof_ids = []
        bc_fext = []
        bc_dof_names = self.bc.dof
        dof_names = self.dof.names

        for element_id, surface_name in self.bc_surface:
            connectivity = elements[element_id]
            node_coords = nodes[connectivity]
            iso_element_type = get_iso_element_type(node_coords)
            iso_element_shape = iso_element_shape_dict[iso_element_type]

            nodes_number = iso_element_shape.nodes_number
            bc_gp_weights = iso_element_shape.bc_gp_weights
            bc_gp_number = len(bc_gp_weights)
            bc_gp_shape_values = iso_element_shape.bc_gp_shape_values_dict[surface_name]
            bc_gp_shape_gradients = iso_element_shape.bc_gp_shape_gradients_dict[surface_name]
            bc_surface_coord = iso_element_shape.bc_surface_coord_dict[surface_name]
            surface_local_nodes = array(iso_element_shape.bc_surface_nodes_dict[surface_name])
            surface_nodes = elements[element_id][surface_local_nodes]
            surface_dof_ids = []
            for node_index in surface_nodes:
                for _, bc_dof_name in enumerate(bc_dof_names):
                    surface_dof_id = node_index * len(dof_names) + dof_names.index(bc_dof_name)
                    surface_dof_ids.append(surface_dof_id)

            dof_ids += surface_dof_ids

            element_fext = zeros(nodes_number)

            for i in range(bc_gp_number):
                bc_gp_jacobi = dot(bc_gp_shape_gradients[i], node_coords).transpose()
                bc_gp_jacobi_sub = delete(bc_gp_jacobi, bc_surface_coord[0], axis=1)
                if dimension == 2:
                    s = sum(bc_gp_jacobi_sub ** 2)
                    element_fext += bc_gp_shape_values[i].transpose() * bc_gp_weights[i] * bc_value * sqrt(s) * \
                                    bc_surface_coord[2]
                elif dimension == 3:
                    s = 0
                    for row in range(bc_gp_jacobi_sub.shape[0]):
                        s += det(delete(bc_gp_jacobi_sub, row, axis=0)) ** 2
                    element_fext += bc_gp_shape_values[i].transpose() * bc_gp_weights[i] * bc_value * sqrt(s) * \
                                    bc_surface_coord[2]

            surface_fext = []
            for fext in element_fext[surface_local_nodes]:
                for _ in range(len(bc_dof_names)):
                    surface_fext.append(fext)

            bc_fext += list(surface_fext)

        self.dof_ids = array(dof_ids)
        self.bc_fext = array(bc_fext)


if __name__ == "__main__":
    from pyfem.io.Properties import Properties

    # props = Properties()
    # props.read_file(r'F:\Github\pyfem\examples\rectangle\rectangle.toml')
    # bc_data = NeumannBCDistributed(props.bcs[3], props.dof, props.mesh_data, props.solver, props.amplitudes[0])
    # bc_data.show()

    # props = Properties()
    # props.read_file(r'F:\Github\pyfem\examples\hex\hex.toml')
    # bc_data = NeumannBCDistributed(props.bcs[4], props.dof, props.mesh_data, props.solver, props.amplitudes[0])
    # bc_data.show()

    props = Properties()
    props.read_file(r'F:\Github\pyfem\examples\quad8\quad8.toml')
    bc_data = NeumannBCDistributed(props.bcs[2], props.dof, props.mesh_data, props.solver, props.amplitudes[0])
    bc_data.show()



# -*- coding: utf-8 -*-
"""

"""
from pathlib import Path
from typing import List, Dict, Union

import meshio  # type: ignore
from numpy import ndarray, empty

from pyfem.fem.constants import DTYPE
from pyfem.utils.colors import BLUE, END
from pyfem.utils.colors import error_style, info_style
from pyfem.utils.wrappers import show_running_time


class MeshData:

    def __init__(self) -> None:
        self.dimension: int = -1
        self.mesh: meshio.Mesh = None  # type: ignore
        self.nodes: ndarray = empty(0)
        self.elements: List[ndarray] = []
        self.bc_elements: List[ndarray] = []
        self.node_sets: Dict[str, List[int]] = {}
        self.element_sets: Dict[str, List[int]] = {}
        self.bc_element_sets: Dict[str, List[int]] = {}

    def show(self) -> None:
        print(self.to_string(0))

    def to_string(self, level: int = 1) -> str:
        msg = BLUE + self.__str__() + END + '\n'
        space = '   ' * level
        msg += space + info_style('  Number of nodes ............ %6d\n' % len(self.nodes))
        if len(self.node_sets) > 0:
            msg += space + info_style('  Number of node_sets ........ %6d\n' % len(self.node_sets))
            msg += space + '  -----------------------------------\n'
            msg += space + '    name                       #nodes\n'
            msg += space + '    ---------------------------------\n'

            for name in self.node_sets:
                msg += space + '    %-16s           %6d\n' % (name, len(self.node_sets[name]))

        msg += '\n'
        msg += space + info_style('  Number of elements ......... %6d\n' % len(self.elements))
        if len(self.element_sets) > 0:
            msg += space + info_style('  Number of element_sets ..... %6d\n' % len(self.element_sets))
            msg += space + '  -----------------------------------\n'
            msg += space + '    name                       #elems\n'
            msg += space + '    ---------------------------------\n'

            for name in self.element_sets:
                msg += space + '    %-16s           %6d\n' % (name, len(self.element_sets[name]))

        msg += '\n'
        msg += space + info_style('  Number of bc_elements ...... %6d\n' % len(self.bc_elements))
        if len(self.bc_element_sets) > 0:
            msg += space + info_style('  Number of bc_element_sets .. %6d\n' % len(self.bc_element_sets))
            msg += space + '  -----------------------------------\n'
            msg += space + '    name                       #elems\n'
            msg += space + '    ---------------------------------\n'

            for name in self.bc_element_sets:
                msg += space + '    %-16s           %6d\n' % (name, len(self.bc_element_sets[name]))

        return msg

    @show_running_time
    def read_file(self, filename: Union[Path, str], file_format: str = "gmsh") -> None:
        self.mesh = meshio.read(filename, file_format)

        # 单元类型和维度的映射
        cell_type_to_dim = {cell.type: cell.dim for cell in self.mesh.cells}

        self.dimension = max(cell_type_to_dim.values())

        # 建立节点坐标数组
        self.nodes = self.mesh.points[:, :self.dimension].astype(DTYPE)

        # 建立实体单元和边界单元列表
        mesh_type_start_number = {}
        start_number = 0
        bc_mesh_type_start_number = {}
        bc_start_number = 0
        elements = []
        bc_elements = []
        for mesh_type, mesh_element_ids in self.mesh.cells_dict.items():
            if cell_type_to_dim[mesh_type] == self.dimension:
                elements += list(mesh_element_ids)
                mesh_type_start_number[mesh_type] = start_number
                start_number += len(mesh_element_ids)
            elif cell_type_to_dim[mesh_type] == self.dimension - 1:
                bc_elements += list(mesh_element_ids)
                bc_mesh_type_start_number[mesh_type] = start_number
                bc_start_number += len(mesh_element_ids)
            else:
                raise NotImplementedError(error_style(f'unsupported dimension {self.dimension}'))
        self.elements = elements
        self.bc_elements = bc_elements

        # 实体单元和边界单元集合
        for cell_set_name, cell_set_dict in self.mesh.cell_sets_dict.items():
            if cell_set_name != 'gmsh:bounding_entities':
                elements_in_set = []
                bc_elements_in_set = []
                for mesh_type, mesh_element_ids in cell_set_dict.items():
                    if cell_type_to_dim[mesh_type] == self.dimension:
                        elements_in_set += list(mesh_element_ids + mesh_type_start_number[mesh_type])
                    elif cell_type_to_dim[mesh_type] == self.dimension - 1:
                        bc_elements_in_set += list(mesh_element_ids + bc_mesh_type_start_number[mesh_type])
                    else:
                        raise NotImplementedError(error_style(f'unsupported dimension {self.dimension}'))
                if elements_in_set:
                    self.element_sets[cell_set_name] = elements_in_set
                if bc_elements_in_set:
                    self.bc_element_sets[cell_set_name] = bc_elements_in_set

        # 基于边界单元集合建立边界节点集合
        for bc_element_set_name, bc_element_ids in self.bc_element_sets.items():
            for bc_element_id in bc_element_ids:
                self.add_to_node_sets(bc_element_set_name, list(bc_elements[bc_element_id]))

        # 去除边界节点集合中重复的节点
        for key in self.node_sets:
            self.node_sets[key] = list(set(self.node_sets[key]))

        # 添加用户定义的节点集合
        for point_set in self.mesh.point_sets:
            self.node_sets[point_set] = list(self.mesh.point_sets[point_set])

    def add_to_node_sets(self, node_set_name: str, node_ids: List[int]) -> None:
        if node_set_name not in self.node_sets:
            self.node_sets[node_set_name] = node_ids
        else:
            self.node_sets[node_set_name] += node_ids


if __name__ == "__main__":
    mesh_data = MeshData()
    # mesh_data.read_file(r'F:\Github\pyfem\examples\rectangle\quad40000.msh', 'gmsh')
    mesh_data.read_file(r'F:\Github\pyfem\examples\hole\hole_quad4.inp', 'abaqus')
    # mesh_data.read_file(r'F:\Github\pyfem\examples\quad_tria\Job-1.inp', 'abaqus')
    mesh_data.show()
